import os
import importlib

# 新版 FeelUOwn 会在启动的时候，设置这个环境变量
qt_moduel_name = os.getenv('FEELUOWN_QT_API', 'PyQt5')

if qt_moduel_name != 'PyQt5':
    qt_moduel_name = 'PyQt6'

QtCore = importlib.import_module(qt_moduel_name + ".QtCore", package=qt_moduel_name)
QtGui = importlib.import_module(qt_moduel_name + ".QtGui", package=qt_moduel_name)
QtWidgets = importlib.import_module(qt_moduel_name + ".QtWidgets", package=qt_moduel_name)

QLineEdit = QtWidgets.QLineEdit
QLabel = QtWidgets.QLabel
QPushButton = QtWidgets.QPushButton
QDialog = QtWidgets.QDialog
QVBoxLayout = QtWidgets.QVBoxLayout
QFileDialog = QtWidgets.QFileDialog
QMessageBox = QtWidgets.QMessageBox
QInputDialog = QtWidgets.QInputDialog
QFrame = QtWidgets.QFrame
QTabWidget = QtWidgets.QTabWidget
QWidget = QtWidgets.QWidget

pyqtSignal = QtCore.pyqtSignal
QThread = QtCore.QThread
QCloseEvent = QtGui.QCloseEvent

if qt_moduel_name == 'PyQt5':
    RichText = QtCore.Qt.RichText
    Password = QLineEdit.Password
    QAction = QtWidgets.QAction
    TextSelectableByMouse = QtCore.Qt.TextSelectableByMouse
    ActionsContextMenu = QtCore.Qt.ActionsContextMenu
    Dialog = QtCore.Qt.Dialog
    TextBrowserInteraction = QtCore.Qt.TextBrowserInteraction
else:
    RichText = QtCore.Qt.TextFormat.RichText
    Password = QLineEdit.EchoMode.Password
    QAction = QtGui.QAction
    TextSelectableByMouse = QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
    ActionsContextMenu = QtCore.Qt.ContextMenuPolicy.ActionsContextMenu
    Dialog = QtCore.Qt.WindowType.Dialog
    TextBrowserInteraction = QtCore.Qt.TextInteractionFlag.TextBrowserInteraction




import logging
from typing import List, Union, Optional, Tuple

from bs4 import BeautifulSoup
from feeluown.library import SongModel, BriefArtistModel, PlaylistModel, BriefPlaylistModel, BriefUserModel, \
    BriefSongModel, ArtistModel, CommentModel, VideoModel, BriefVideoModel, BriefAlbumModel, AlbumModel, \
    SimpleSearchResult

from fuo_bilibili import __identifier__
from fuo_bilibili.api import SearchType
from fuo_bilibili.api.schema.requests import SearchRequest
from fuo_bilibili.api.schema.responses import SearchResponse, SearchResultVideo, VideoInfoResponse, \
    FavoriteListResponse, FavoriteInfoResponse, FavoriteResourceResponse, CollectedFavoriteListResponse, \
    FavoriteSeasonResourceResponse, HistoryLaterVideoResponse, HomeDynamicVideoResponse, UserInfoResponse, \
    UserBestVideoResponse, UserVideoResponse, AudioFavoriteSongsResponse, AudioFavoriteListResponse, AudioPlaylist, \
    AudioPlaylistSong, VideoHotCommentsResponse, SearchResultUser, LiveFeedListResponse, SearchResultLiveRoom, \
    SearchResultMedia, MediaGetListResponse, VideoWeeklyListResponse, VideoMostPopularResponse, \
    HomeRecommendVideosResponse
from fuo_bilibili.util import format_timedelta_to_hms

logger = logging.getLogger(__name__)
PROVIDER_ID = __identifier__


def get_text_from_html(html):
    return BeautifulSoup(html, features="html.parser").get_text()


def wrap_pic(pic):
    if pic.startswith('//'):
        return f'http:{pic}'
    return pic


class BBriefAlbumModel(BriefAlbumModel):
    cover: str = ''


class BSongModel(SongModel):
    source: str = PROVIDER_ID

    lyric: str = None

    @classmethod
    def create_audio_model(cls, au: AudioPlaylistSong) -> Optional['BSongModel']:
        return BriefSongModel(
            source=__identifier__,
            identifier=f'audio_{au.id}',
            title=au.title,
            artists_name=au.author,
            duration_ms=format_timedelta_to_hms(au.duration),
        )

    @classmethod
    def create_user_brief_model(cls, media: UserVideoResponse.UserVideoResponseData.RList.Video):
        return BriefSongModel(
            source=__identifier__,
            identifier=media.bvid,
            title=media.title,
            artists_name=media.author,
            duration_ms=media.length
        )

    @classmethod
    def create_hot_model(cls, result: UserBestVideoResponse.BestVideo):
        return BriefSongModel(
            source=__identifier__,
            identifier=result.bvid,
            title=BeautifulSoup(result.title).get_text(),
            artists_name=result.owner.name,
            duration_ms=format_timedelta_to_hms(result.duration)
        )

    @classmethod
    def create_dynamic_brief_model(cls, item: HomeDynamicVideoResponse.HomeDynamicVideoResponseData.DynamicVideoItem):
        media = item.modules.module_dynamic.major.archive
        author = item.modules.module_author
        return BriefSongModel(
            source=__identifier__,
            identifier=media.bvid,
            title=media.title,
            artists_name=author.name,
            duration_ms=media.duration_text,
        )

    @classmethod
    def create_brief_model(cls, media: FavoriteResourceResponse.FavoriteResourceResponseData.Media) -> BriefSongModel:
        return BriefSongModel(
            source=__identifier__,
            identifier=media.bvid,
            title=media.title,
            artists_name=media.upper.name,
            duration_ms=format_timedelta_to_hms(media.duration)
        )

    @classmethod
    def create_model(cls, result: SearchResultVideo) -> 'BSongModel':
        return BriefSongModel(
            source=__identifier__,
            identifier=result.bvid,
            title=BeautifulSoup(result.title).get_text(),
            artists_name=result.author,
            duration_ms=format_timedelta_to_hms(result.duration)
        )

    @classmethod
    def create_info_model(cls, response: VideoInfoResponse) -> 'BSongModel':
        result = response.data
        lrc = None
        if result.subtitle.list is not None and len(result.subtitle.list) > 0:
            lrc = result.subtitle.list[0].subtitle_url

        children = []
        for page in result.pages:
            # Need not set album and artists for song.children,
            # because they are actually medias with titles and identifier.
            song = cls(
                source=__identifier__,
                identifier=f'paged_{result.bvid}__{page.page}',
                album=None,
                title=page.part,
                artists=[],
                duration=page.duration.total_seconds() * 1000,
                pic_url=wrap_pic(result.pic),
            )
            children.append(song)

        return cls(
            source=__identifier__,
            identifier=result.bvid,
            album=BBriefAlbumModel(
                source=PROVIDER_ID,
                identifier=result.owner.mid,
                name=result.owner.name,
                artists_name=result.owner.name,
                cover=result.owner.face,
            ),
            title=result.title,
            artists=[BriefArtistModel(
                source=PROVIDER_ID,
                identifier=result.owner.mid,
                name=result.owner.name,
            )],
            duration=result.duration.total_seconds() * 1000,
            lyric=lrc or '',
            pic_url=wrap_pic(result.pic),
            children=children,
        )

    @classmethod
    def create_history_brief_model(cls, media):
        return BriefSongModel(
            source=__identifier__,
            identifier=media.bvid,
            title=media.title,
            artists_name=media.owner.name,
            duration_ms=format_timedelta_to_hms(media.duration)
        )

    @classmethod
    def create_history_brief_model_list(cls, resp: HistoryLaterVideoResponse) -> List[BriefSongModel]:
        if resp.data.list is None:
            return []
        return [cls.create_history_brief_model(media) for media in resp.data.list]


class BAlbumModel(AlbumModel):
    @staticmethod
    def create_episode_song_model(ep: MediaGetListResponse.MediaGetListResponseData.Episode,
                                  media: MediaGetListResponse.MediaGetListResponseData) -> BSongModel:
        return BSongModel(
            source=__identifier__,
            identifier=ep.bvid,
            album=BBriefAlbumModel(
                source=PROVIDER_ID,
                identifier=f'media_{media.media_id}_{media.season_id}',
                name=BeautifulSoup(media.title).text,
                artists_name='',
                cover=media.cover,
            ),
            title=ep.long_title,
            artists=[],
            duration=ep.duration.total_seconds(),
        )

    @classmethod
    def create_season_model(cls, media: MediaGetListResponse.MediaGetListResponseData) -> 'BAlbumModel':
        return cls(
            source=PROVIDER_ID,
            identifier=f'media_{media.media_id}_{media.season_id}',
            name=BeautifulSoup(media.title).text,
            artists=[],
            songs=[cls.create_episode_song_model(ep, media) for ep in media.episodes],
            cover=media.cover.replace('http://', 'https://'),
            description=media.evaluate,
        )


class BSearchModel:
    # ['q', 'songs', 'playlists', 'artists', 'albums', 'videos']
    q: str
    songs = List[BSongModel]

    @staticmethod
    def search_user_model(user: SearchResultUser):
        return BriefArtistModel(
            source=PROVIDER_ID,
            identifier=user.mid,
            name=user.uname,
        )

    @staticmethod
    def search_media_model(media: SearchResultMedia):
        return BBriefAlbumModel(
            source=PROVIDER_ID,
            identifier=f'media_{media.media_id}_{media.season_id}',
            name=BeautifulSoup(media.title).text,
            artists_name='',
            cover=media.cover.replace('http://', 'https://'),
        )

    @staticmethod
    def search_live_model(live: SearchResultLiveRoom):
        return BVideoModel(
            source=PROVIDER_ID,
            identifier=f'live_{live.roomid}',
            title=get_text_from_html(live.title),
            artists=[
                BriefArtistModel(
                    source=PROVIDER_ID,
                    identifier=live.uid,
                    name=live.uname
                )
            ],
            duration=0,
            cover=f'https:{live.user_cover}',
        )

    @classmethod
    def create_model(cls, request: SearchRequest, response: Union[SearchResponse, Tuple[SearchResponse]]):
        songs = []
        artists = []
        videos = []
        albums = []
        invalid_count = 0
        if isinstance(response, Tuple):
            results = []
            for r in response:
                if r.data.result is None:
                    continue
                results += r.data.result
        else:
            results = response.data.result or []
        # fixme: implements multiple search types
        for result in results:
            if isinstance(result, SearchResultVideo):
                songs.append(BSongModel.create_model(result))
                videos.append(BVideoModel.create_video_model(result))
            elif isinstance(result, SearchResultUser):
                artists.append(cls.search_user_model(result))
            elif isinstance(result, SearchResultLiveRoom):
                videos.append(cls.search_live_model(result))
            elif isinstance(result, SearchResultMedia):
                albums.append(cls.search_media_model(result))
            else:
                invalid_count += 1
        if invalid_count != 0:
            logger.warning(f'invalid object in search result: {invalid_count}')

        return SimpleSearchResult(
            q=request[0].keyword if isinstance(request, Tuple) else request.keyword,
            songs=songs or [],
            artists=artists or [],
            videos=videos or [],
            albums=albums or [],
        )


class BPlaylistModel(PlaylistModel):
    count: int

    @classmethod
    def create_audio_model(cls, p: AudioPlaylist):
        identifier = None
        desc = ''
        count = 0
        match p.type:
            case 1:
                identifier = f'audio_{p.type}_{p.id}'
                count = p.song or 0
                desc = p.desc
            case 2:
                identifier = f'audio_{p.type}_{p.menuId}'
                count = p.snum or 0
                desc = p.intro
        if identifier is None:
            return None
        return cls(
            source=PROVIDER_ID,
            identifier=identifier,
            creator=BriefUserModel(
                source=PROVIDER_ID,
                identifier=p.uid,
                name=p.uname
            ),
            name=f'{p.title} (音频)',
            cover='',
            description=desc,
            count=count,
        )

    @classmethod
    def create_audio_brief_model(cls, p: AudioPlaylist):
        identifier = None
        match p.type:
            case 1:
                identifier = f'audio_{p.type}_{p.id}'
            case 2:
                identifier = f'audio_{p.type}_{p.menuId}'
        if identifier is None:
            return None
        return BriefPlaylistModel(
            source=PROVIDER_ID,
            identifier=identifier,
            creator_name=p.uname,
            name=f'{p.title} (音频)',
        )

    @classmethod
    def weekly_brief_model(cls, item: VideoWeeklyListResponse.VideoWeeklyListResponseData.VideoWeeklyItem):
        return BriefPlaylistModel(
            source=PROVIDER_ID,
            identifier=f'weekly_{item.number}',
            creator_name='',
            name=f'{item.subject} ({item.name})',
        )

    @classmethod
    def create_audio_model_list(cls, resp: AudioFavoriteListResponse) -> List[BriefPlaylistModel]:
        if resp.data is None:
            return []
        return [cls.create_audio_brief_model(p) for p in resp.data.data]

    @classmethod
    def create_brief_model(cls, fav: Union[FavoriteListResponse.FavoriteListResponseData.FavoriteList,
                                           CollectedFavoriteListResponse.CollectedFavoriteListResponseData
                           .CollectedFavoriteList]):
        return BriefPlaylistModel(
            source=PROVIDER_ID,
            identifier=f'{fav.type}_{fav.id}',
            creator_name='',
            name=fav.title,
        )

    @classmethod
    def special_brief_playlists(cls):
        return [
            BriefPlaylistModel(
                source=PROVIDER_ID,
                identifier=f'DYNAMIC',
                creator_name='',
                name='动态视频',
            ),
            BriefPlaylistModel(
                source=PROVIDER_ID,
                identifier=f'LATER',
                creator_name='',
                name='稍后再看',
            ),
            BriefPlaylistModel(
                source=PROVIDER_ID,
                identifier=f'HISTORY',
                creator_name='',
                name='历史记录',
            ),
        ]

    @classmethod
    def special_model(cls, identifier, resp: Optional[HistoryLaterVideoResponse]):
        match identifier:
            case 'DYNAMIC':
                return cls(
                    source=PROVIDER_ID,
                    identifier=f'DYNAMIC',
                    creator=None,
                    name='动态视频',
                    cover='',
                    description='动态视频',
                    count=1000,
                )
            case 'LATER':
                return cls(
                    source=PROVIDER_ID,
                    identifier=f'LATER',
                    creator=None,
                    name='稍后再看',
                    cover='',
                    description='稍后再看',
                    count=resp.data.count if resp is not None else None,
                )
            case 'HISTORY':
                return cls(
                    source=PROVIDER_ID,
                    identifier=f'HISTORY',
                    creator=None,
                    name='历史纪录',
                    cover='',
                    description='历史纪录',
                    count=1000,
                )

    @classmethod
    def create_info_model(cls, response: Union[FavoriteInfoResponse, FavoriteSeasonResourceResponse]):
        if isinstance(response, FavoriteSeasonResourceResponse):
            return cls(
                source=PROVIDER_ID,
                identifier=f'21_{response.data.info.id}',
                creator=BriefUserModel(
                    source=PROVIDER_ID,
                    identifier=response.data.info.upper.mid,
                    name=response.data.info.upper.name,
                ),
                name=response.data.info.title,
                cover=response.data.info.cover,
                description=f'{response.data.info.media_count}个视频，{response.data.info.cnt_info.play}播放',
                count=response.data.info.media_count,
            )
        else:
            return cls(
                source=PROVIDER_ID,
                identifier=f'{response.data.type}_{response.data.id}',
                creator=BriefUserModel(
                    source=PROVIDER_ID,
                    identifier=response.data.upper.mid,
                    name=response.data.upper.name,
                ),
                name=response.data.title,
                cover=response.data.cover,
                description=response.data.intro,
                count=response.data.media_count,
            )

    @classmethod
    def create_model_list(cls, response: Union[FavoriteListResponse, CollectedFavoriteListResponse]):
        return [cls.create_brief_model(f) for f in response.data.list]


class BArtistModel(ArtistModel):
    @classmethod
    def create_model(cls, resp: UserInfoResponse, video_resp: UserBestVideoResponse) -> 'BArtistModel':
        alias = []
        if resp.data.fans_badge and resp.data.fans_medal.show and resp.data.fans_medal.wear \
                and resp.data.fans_medal.medal is not None:
            alias.append(f'Lv{resp.data.fans_medal.medal.level} {resp.data.fans_medal.medal.medal_name}')
        return cls(
            source=PROVIDER_ID,
            identifier=resp.data.mid,
            name=resp.data.name,
            pic_url=resp.data.face,
            aliases=alias,
            hot_songs=[BSongModel.create_hot_model(v) for v in video_resp.data],
            description=resp.data.sign
        )


class BCommentModel(CommentModel):
    @classmethod
    def create_model(cls, reply: VideoHotCommentsResponse.VideoHotCommentsResponseData.Reply):
        return cls(
            source=PROVIDER_ID,
            identifier=str(reply.rpid),
            user=BriefUserModel(
                source=PROVIDER_ID,
                identifier=reply.member.mid,
                name=reply.member.uname
            ),
            content=reply.content.message,
            liked_count=reply.like,
            time=int(reply.ctime.timestamp())
        )


class BVideoModel(VideoModel):
    @classmethod
    def create_live_model(cls, live: LiveFeedListResponse.LiveFeedListResponseData.LiveFeed):
        return cls(
            source=PROVIDER_ID,
            identifier=f'live_{live.roomid}',
            title=get_text_from_html(live.title),
            artists=[
                BriefArtistModel(
                    source=PROVIDER_ID,
                    identifier=live.uid,
                    name=live.uname
                )
            ],
            duration=0,
            cover=live.cover,
        )

    @classmethod
    def create_recommend_video_model(cls, result: HomeRecommendVideosResponse.HomeRecommendVideosResponseData.Video):
        return VideoModel(
            source=PROVIDER_ID,
            identifier=result.bvid,
            title=get_text_from_html(result.title),
            artists=[BriefArtistModel(
                source=PROVIDER_ID,
                identifier=result.owner.mid,
                name=result.owner.name
            )],
            duration=result.duration.total_seconds(),
            cover=wrap_pic(result.pic),
            play_count=result.stat.view,
        )


    @classmethod
    def create_popular_video_model(cls, result: VideoMostPopularResponse.VideoMostPopularResponseData.PopularVideo):
        return VideoModel(
            source=PROVIDER_ID,
            identifier=result.bvid,
            title=get_text_from_html(result.title),
            artists=[BriefArtistModel(
                source=PROVIDER_ID,
                identifier=result.owner.mid,
                name=result.owner.name
            )],
            duration=result.duration.total_seconds(),
            cover=wrap_pic(result.pic),
            play_count=result.stat.view,
            released=result.pubdate.strftime('%Y-%m-%d'),
        )

    @classmethod
    def create_video_model(cls, result: SearchResultVideo) -> 'VideoModel':
        return VideoModel(
            source=__identifier__,
            identifier=result.bvid,
            title=get_text_from_html(result.title),
            artists=[BriefArtistModel(
                source=__identifier__,
                name=result.author,
                identifier=result.mid
            )],
            duration=result.duration.total_seconds(),
            cover=wrap_pic(result.pic),
            play_count=result.play,
            released=result.pubdate.strftime('%Y-%m-%d'),
        )




from typing import Type, Any, Optional, Union, List

from .compat import BaseModel

from fuo_bilibili.api.schema.requests import BaseRequest, FavoriteListRequest, FavoriteInfoRequest, \
    FavoriteResourceRequest, CollectedFavoriteListRequest, FavoriteSeasonResourceRequest, \
    FavoriteResourceOperateRequest, FavoriteNewRequest
from fuo_bilibili.api.schema.responses import BaseResponse, FavoriteListResponse, FavoriteInfoResponse, \
    FavoriteResourceResponse, CollectedFavoriteListResponse, FavoriteSeasonResourceResponse


class PlaylistMixin:
    """收藏夹接口"""
    APIX_BASE = 'https://api.bilibili.com/x'

    def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None]) -> Any:
        pass

    def post(self, url: str, param: Optional[BaseRequest], clazz: Type[BaseResponse], is_json=False, **kwargs) -> Any:
        pass

    @staticmethod
    def clear_cache_by_url(urls: List[str]):
        pass

    def favorite_list(self, request: FavoriteListRequest) -> FavoriteListResponse:
        url = f'{self.APIX_BASE}/v3/fav/folder/created/list-all'
        return self.get(url, request, FavoriteListResponse)

    def collected_favorite_list(self, request: CollectedFavoriteListRequest) -> CollectedFavoriteListResponse:
        url = f'{self.APIX_BASE}/v3/fav/folder/collected/list'
        return self.get(url, request, CollectedFavoriteListResponse)

    def favorite_info(self, request: FavoriteInfoRequest) -> FavoriteInfoResponse:
        url = f'{self.APIX_BASE}/v3/fav/folder/info'
        return self.get(url, request, FavoriteInfoResponse)

    def favorite_resource(self, request: FavoriteResourceRequest) -> FavoriteResourceResponse:
        url = f'{self.APIX_BASE}/v3/fav/resource/list'
        return self.get(url, request, FavoriteResourceResponse)

    def favorite_season_resource(self, request: FavoriteSeasonResourceRequest) -> FavoriteSeasonResourceResponse:
        url = f'{self.APIX_BASE}/space/fav/season/list'
        return self.get(url, request, FavoriteSeasonResourceResponse)

    def favorite_resource_operate(self, request: FavoriteResourceOperateRequest) -> BaseResponse:
        url = f'{self.APIX_BASE}/v3/fav/resource/deal'
        return self.post(url, request, BaseResponse)

    def favorite_new(self, request: FavoriteNewRequest) -> BaseResponse:
        url = f'{self.APIX_BASE}/v3/fav/folder/add'
        self.clear_cache_by_url([f'{self.APIX_BASE}/v3/fav/folder/created/list-all'])
        return self.post(url, request, BaseResponse)



from typing import Type, Any, Optional, Union

from .compat import BaseModel

from fuo_bilibili.api.schema.requests import BaseRequest, PasswordLoginRequest, SendSmsCodeRequest, \
    SmsCodeLoginRequest, BaseCsrfRequest
from fuo_bilibili.api.schema.responses import BaseResponse, RequestCaptchaResponse, \
    RequestLoginKeyResponse, PasswordLoginResponse, SendSmsCodeResponse, SmsCodeLoginResponse, \
    CookieInfoResponse


class LoginMixin:
    API_BASE = 'https://api.bilibili.com/x/web-interface'
    PASSPORT_BASE = 'https://passport.bilibili.com'

    def get_uncached(self, url: str, param: Optional[BaseRequest],
                     clazz: Union[Type[BaseResponse], Type[BaseModel], None]) -> Any:
        pass

    def get(self, url: str, param: Optional[BaseRequest],
            clazz: Union[Type[BaseResponse], Type[BaseModel], None]) -> Any:
        pass

    def post(self, url: str, param: Optional[BaseRequest], clazz: Type[BaseResponse], is_json=False, **kwargs) -> Any:
        pass

    def _dump_cookie_to_file(self):
        pass

    def cookie_info(self) -> CookieInfoResponse:
        url = f'{self.PASSPORT_BASE}/x/passport-login/web/cookie/info'
        return self.get_uncached(url, BaseCsrfRequest(), CookieInfoResponse)

    def request_captcha(self) -> RequestCaptchaResponse:
        url = f'{self.PASSPORT_BASE}/x/passport-login/captcha'
        return self.get_uncached(url, None, RequestCaptchaResponse)

    def request_login_key(self) -> RequestLoginKeyResponse:
        url = f'{self.PASSPORT_BASE}/x/passport-login/web/key'
        response: RequestLoginKeyResponse = self.get_uncached(url, None, RequestLoginKeyResponse)
        # 接口没有状态码 pydantic 判断字段存在 根据字段是否空判断
        if response.key == '' or response.hash == '':
            raise RuntimeError('getkey error')
        return response

    def send_sms_code(self, request: SendSmsCodeRequest) -> SendSmsCodeResponse:
        url = f'{self.PASSPORT_BASE}/x/passport-login/web/sms/send'
        return self.post(url, request, SendSmsCodeResponse)

    def sms_code_login(self, request: SmsCodeLoginRequest) -> SmsCodeLoginResponse:
        url = f'{self.PASSPORT_BASE}/x/passport-login/web/login/sms'
        resp: SmsCodeLoginResponse = self.post(url, request, SmsCodeLoginResponse)
        if resp.data.url is not None:
            self.get_uncached(resp.data.url, None, None)
        self._dump_cookie_to_file()
        return resp

    def password_login(self, request: PasswordLoginRequest) -> PasswordLoginResponse:
        """
        账号密码登录
        """
        url = f'{self.PASSPORT_BASE}/x/passport-login/web/login'
        resp: PasswordLoginResponse = self.post(url, request, PasswordLoginResponse,
                                                headers={
                                                    'user-agent': 'Mozilla/5.0',
                                                    'referer': 'https://passport.bilibili.com/login',
                                                    'content-type': 'application/x-www-form-urlencoded'
                                                })
        if resp.data.status != 2:
            self._dump_cookie_to_file()
        return resp



from fuo_bilibili.api.compat import BaseModel, Field

from fuo_bilibili.api.schema.enums import VideoQualityNum, VideoFnval, SearchType, SearchOrderType, UserType, \
    VideoDurationType, FavoriteResourceOrderType, CommentType


class BaseRequest(BaseModel):
    def __hash__(self):
        return hash(self.json())


class BaseWbiRequest(BaseRequest):
    pass


class BaseOptionalWbiRequest(BaseRequest):
    pass


class GeetestBase(BaseModel):
    token: str
    challenge: str
    validate_: str = Field(alias='validate')
    seccode: str


class PaginatedRequest(BaseRequest):
    pn: int = 1  # 页码
    ps: int = 20  # 每页数量


class AnotherPaginatedRequest(BaseRequest):
    page: int = 1
    pagesize: int = 20


class SendSmsCodeRequest(BaseRequest, GeetestBase):
    tel: str  # 手机号（不包含国家代码）
    cid: str = '+86'  # 国家代码
    source: str = 'main_web'


class SmsCodeLoginRequest(BaseRequest):
    tel: str  # 手机号（不包含国家代码）
    cid: str = '+86'  # 国家代码
    code: str  # 验证码
    source: str = 'main_web'
    captcha_key: str
    keep: bool = True


class PasswordLoginRequest(BaseRequest, GeetestBase):
    source: str = 'main_h5'
    username: str
    password: str  # 加密后密码
    keep: bool = True


class SearchRequest(BaseRequest):
    search_type: SearchType
    keyword: str
    order: SearchOrderType = None
    order_sort: int = None  # 用户粉丝数及等级排序顺序 1:升序 0:降序(default)
    user_type: UserType = None
    duration: VideoDurationType = None
    tids: int = None  # 视频分区
    category_id: int = None  # 专栏分区 todo enum
    page: int = None  # 分页


class VideoInfoRequest(BaseRequest):
    aid: int = None
    bvid: str


class PlayUrlRequest(BaseOptionalWbiRequest):
    avid: int = None
    bvid: str
    cid: int
    qn: VideoQualityNum = None
    fnval: VideoFnval = None
    fnver: int = 0
    fourk: int = None  # 是否允许4K视频
    otype: str = 'json'
    platform: str = 'pc'


class FavoriteListRequest(BaseRequest):
    up_mid: int  # UID
    type: int = 2  # 2:视频
    rid: int = None  # 资源ID


class CollectedFavoriteListRequest(PaginatedRequest):
    up_mid: int  # UID
    platform: str = 'web'


class FavoriteInfoRequest(BaseRequest):
    media_id: int  # 收藏夹ID


class FavoriteResourceRequest(PaginatedRequest):
    media_id: int  # 收藏夹ID
    keyword: str = None  # 关键词搜索
    order: FavoriteResourceOrderType = FavoriteResourceOrderType.MTIME  # 排序方式
    type: int = 0
    tid: int = 0


class FavoriteSeasonResourceRequest(PaginatedRequest):
    season_id: int  # 合集ID


class HomeRecommendVideosRequest(PaginatedRequest):
    refresh_type: int = 3
    version: int = 1
    fresh_idx: int = 2
    fresh_idx_1h: int = 2
    homepage_ver: int = 1


class HomeDynamicVideoRequest(BaseRequest):
    timezone_offset: int = -480
    type: str = 'video'
    offset: str = None
    page: int = 1


class UserInfoRequest(BaseWbiRequest):
    mid: int  # 用户UID


class UserBestVideoRequest(BaseRequest):
    vmid: int


class UserVideoRequest(BaseWbiRequest, PaginatedRequest):
    mid: int
    order: str = None
    tid: int = None
    keyword: str = None


class AudioFavoriteSongsRequest(PaginatedRequest):
    sid: int  # 音频收藏夹mlid


class AudioGetUrlRequest(BaseRequest):
    sid: int
    privilege: int = 2
    quality: int = 2


class VideoHotCommentsRequest(PaginatedRequest):
    type: CommentType = CommentType.video
    oid: int
    root: int = None
    sort: int = 2  # 2:热度 0:时间


class LivePlayUrlRequest(BaseRequest):
    cid: int  # 直播间roomid
    platform: str = 'h5'
    # 80：流畅
    # 150：高清
    # 400：蓝光
    # 10000：原画
    # 20000：4K
    # 30000：杜比
    qn: int = 10000


class MediaGetListRequest(BaseRequest):
    season_id: int = None
    ep_id: int = None


class MediaFavlistRequest(PaginatedRequest):
    vmid: int
    type: int = 1


class BaseCsrfRequest(BaseRequest):
    csrf: str = ''


class HistoryAddLaterVideosRequest(BaseCsrfRequest):
    bvid: str


class HistoryDelLaterVideosRequest(BaseCsrfRequest):
    aid: int


class FavoriteResourceOperateRequest(BaseCsrfRequest):
    rid: int
    type: int = 2
    add_media_ids: str = ''
    del_media_ids: str = ''


class FavoriteNewRequest(BaseCsrfRequest):
    title: str
    intro: str
    privacy: int
    cover: str = ''


class UserFollowingRequest(PaginatedRequest):
    vmid: int
    order: str = 'desc'


class WeeklyDetailRequest(BaseRequest):
    number: int



import logging
import json

from .const import PLUGIN_API_COOKIEDICT_FILE


logger = logging.getLogger(__name__)


def load_user_cookies():
    """
    Load cookies from a file in JSON format.
    """
    if PLUGIN_API_COOKIEDICT_FILE.exists():
        with PLUGIN_API_COOKIEDICT_FILE.open('r', encoding='utf-8') as f:
            try:
                cookie_dict = json.load(f)
            except Exception:
                logger.warning('parse cookies(json) failed')
                return None
        return cookie_dict


def dump_user_cookies(cookies):
    """
    Dump cookies to a file in JSON format.
    """
    with PLUGIN_API_COOKIEDICT_FILE.open('w', encoding='utf-8') as f:
        json.dump(cookies, f, indent=2)



import logging
import math
from inspect import signature
from pathlib import Path
from time import time
from typing import List, Optional, Union, Tuple

from feeluown.excs import NoUserLoggedIn, MediaNotFound
from feeluown.library import AbstractProvider, ProviderV2, ProviderFlags as Pf, UserModel, VideoModel, \
    BriefPlaylistModel, BriefSongModel, LyricModel, SupportsSongSimilar, BriefSongProtocol, SupportsSongHotComments, \
    SupportsAlbumGet, BriefAlbumModel, SupportsPlaylistAddSong, SupportsPlaylistRemoveSong, BriefUserModel, \
    BriefArtistModel, SearchType as FuoSearchType, ModelType, SimpleSearchResult, ModelNotFound, Collection, \
    CollectionType
from feeluown.media import Quality, Media, MediaType, VideoAudioManifest
from feeluown.utils.reader import SequentialReader

from fuo_bilibili import __identifier__, __alias__
from fuo_bilibili.api import BilibiliApi, SearchRequest, SearchType as BilibiliSearchType, VideoInfoRequest, \
    PlayUrlRequest, VideoQualityNum
from fuo_bilibili.api.schema.enums import VideoFnval, CommentType
from fuo_bilibili.api.schema.requests import PasswordLoginRequest, SendSmsCodeRequest, SmsCodeLoginRequest, \
    FavoriteListRequest, FavoriteInfoRequest, FavoriteResourceRequest, CollectedFavoriteListRequest, \
    FavoriteSeasonResourceRequest, PaginatedRequest, HomeRecommendVideosRequest, HomeDynamicVideoRequest, \
    UserInfoRequest, UserBestVideoRequest, UserVideoRequest, AudioFavoriteSongsRequest, AudioGetUrlRequest, \
    VideoHotCommentsRequest, AnotherPaginatedRequest, LivePlayUrlRequest, MediaGetListRequest, MediaFavlistRequest, \
    HistoryAddLaterVideosRequest, HistoryDelLaterVideosRequest, FavoriteResourceOperateRequest, FavoriteNewRequest, \
    UserFollowingRequest, WeeklyDetailRequest
from fuo_bilibili.api.schema.responses import RequestCaptchaResponse, RequestLoginKeyResponse, PasswordLoginResponse, \
    SendSmsCodeResponse, SmsCodeLoginResponse, NavInfoResponse, PlayUrlResponse
from fuo_bilibili.const import DANMAKU_DIRECTORY
from fuo_bilibili.danmaku2ass import Danmaku2ASS
from fuo_bilibili.model import BSearchModel, BSongModel, BPlaylistModel, BArtistModel, BCommentModel, BVideoModel, \
    BAlbumModel
from fuo_bilibili.util import json_to_lrc_text
from fuo_bilibili.login import load_user_cookies

logger = logging.getLogger(__name__)
SEARCH_TYPE_MAP = {
    # 对应直播间
    FuoSearchType.vi: BilibiliSearchType.LIVE_ROOM,
    FuoSearchType.ar: BilibiliSearchType.BILI_USER,  # 对应B站用户
    FuoSearchType.so: BilibiliSearchType.VIDEO,  # 对应投稿视频
    FuoSearchType.al: (BilibiliSearchType.MEDIA, BilibiliSearchType.BANGUMI),  # 对应番剧电影
}


def is_child_model(model):
    return model.identifier.startswith('paged_')


def parse_paged_identifier(identifier: str):
    identifier, page_num = identifier[6:].split('__')
    return identifier, page_num


class BilibiliProvider(AbstractProvider, ProviderV2, SupportsSongSimilar, SupportsSongHotComments, SupportsAlbumGet,
                       SupportsPlaylistAddSong, SupportsPlaylistRemoveSong):
    def playlist_add_song(self, playlist, song) -> bool:
        if is_child_model(song):
            return False

        match playlist.identifier:
            case 'LATER':
                self._api.history_add_later_videos(HistoryAddLaterVideosRequest(bvid=song.identifier))
                return True
            case 'HISTORY':
                return False
        if playlist.identifier.startswith('audio_') or song.identifier.startswith('audio_'):
            return False
        _, id_ = playlist.identifier.split('_')
        self._api.favorite_resource_operate(FavoriteResourceOperateRequest(
            rid=self._get_video_avid(song.identifier),
            add_media_ids=str(id_)
        ))
        return True

    def playlist_remove_song(self, playlist, song) -> bool:
        if is_child_model(song):
            return False

        match playlist.identifier:
            case 'LATER':
                self._api.history_del_later_videos(
                    HistoryDelLaterVideosRequest(aid=self._get_video_avid(song.identifier)))
                return True
            case 'HISTORY':
                return False
        if playlist.identifier.startswith('audio_') or song.identifier.startswith('audio_'):
            return False
        self._api.favorite_resource_operate(FavoriteResourceOperateRequest(
            rid=self._get_video_avid(song.identifier),
            del_media_ids=str(playlist.identifier)
        ))
        return True

    # noinspection PyPep8Naming
    class meta:
        identifier: str = __identifier__
        name: str = __alias__
        flags: dict = {
            ModelType.song: (Pf.model_v2 | Pf.get | Pf.multi_quality | Pf.lyric | Pf.mv | Pf.similar | Pf.hot_comments),
            ModelType.video: (Pf.model_v2 | Pf.multi_quality | Pf.get),
            ModelType.playlist: (Pf.model_v2 | Pf.get | Pf.songs_rd | Pf.add_song | Pf.remove_song),
            ModelType.artist: (Pf.model_v2 | Pf.get | Pf.songs_rd),
            ModelType.album: (Pf.model_v2 | Pf.get),
        }

    def __init__(self, enable_live_room_as_video=True):
        super(BilibiliProvider, self).__init__()
        self._api = BilibiliApi()
        self._user = None
        self._video_quality_codes = dict()
        self._video_cids = dict()
        self._video_avids = dict()

        self.enable_live_room_as_video = enable_live_room_as_video

    def _format_search_request(self, keyword, type_) -> Union[SearchRequest, Tuple[SearchRequest]]:
        if self.enable_live_room_as_video is False and type_ == FuoSearchType.vi:
            btype = BilibiliSearchType.VIDEO
        else:
            btype = SEARCH_TYPE_MAP.get(type_)
        if btype is None:
            raise NotImplementedError
        if isinstance(btype, Tuple):
            return tuple([SearchRequest(search_type=t, keyword=keyword) for t in btype])
        return SearchRequest(search_type=btype, keyword=keyword)

    def request_captcha(self) -> RequestCaptchaResponse.RequestCaptchaResponseData:
        res = self._api.request_captcha()
        assert hasattr(res.data, 'geetest') and res.data.geetest is not None
        return res.data

    def request_key(self) -> RequestLoginKeyResponse:
        return self._api.request_login_key()

    def auto_login(self):
        logger.info("Try auto logging...")
        cookiedict = load_user_cookies()
        if cookiedict:
            logger.info("Try fetch user info with cookies...")
            self._api.from_cookiedict(cookiedict)
            try:
                user = self.user_info()
            except RuntimeError as e:
                logger.warning(f'Auto login failed: {e}')
            else:
                self.auth(user)
                logger.info(f'Auto login success: {user.name}')
        else:
            logger.info('No cookies found')

    def auth(self, user):
        self._api.enable_auto_save_cookie = bool(user)
        self._user = user

    def has_current_user(self) -> bool:
        return self._user is not None

    def get_current_user(self):
        if self._user is None:
            raise NoUserLoggedIn
        return self._user

    def user_info(self) -> UserModel:
        data: NavInfoResponse.NavInfoResponseData = self._api.nav_info().data
        self._api.set_wbi(data.wbi_img)
        user = UserModel(
            source=__identifier__,
            identifier=str(data.mid),
            name=data.uname,
            avatar_url=data.face
        )
        return user

    def cookiejar_login(self, jar):
        self._api.from_cookiejar(jar)

    def sms_send_code(self, request: SendSmsCodeRequest) -> SendSmsCodeResponse:
        return self._api.send_sms_code(request)

    def password_login(self, request: PasswordLoginRequest) -> PasswordLoginResponse:
        return self._api.password_login(request)

    def sms_code_login(self, request: SmsCodeLoginRequest) -> SmsCodeLoginResponse:
        return self._api.sms_code_login(request)

    def search(self, keyword, type_, *args, **kwargs) -> Optional[SimpleSearchResult]:
        if not self.has_current_user():
            return None
        request = self._format_search_request(keyword, type_)
        if isinstance(request, Tuple):
            response = tuple([self._api.search(r) for r in request])
        else:
            response = self._api.search(request)
        return BSearchModel.create_model(request, response)

    def song_get(self, identifier) -> Optional[BSongModel]:
        if identifier.startswith('audio_'):
            return None

        if identifier.startswith('paged_'):
            identifier, page_num = parse_paged_identifier(identifier)
        else:
            identifier, page_num = identifier, None

        response = self._api.video_get_info(VideoInfoRequest(bvid=identifier))
        song = BSongModel.create_info_model(response)
        if page_num is not None:
            try:
                song.children[int(page_num) - 1]
            except IndexError:
                return None
        return song

    def song_list_similar(self, song: BriefSongProtocol) -> List[BriefSongModel]:
        if song.identifier.startswith('audio_'):
            return []
        if is_child_model(song):
            return []
        resp = self._api.video_get_related(VideoInfoRequest(bvid=song.identifier))
        return [BSongModel.create_history_brief_model(media) for media in resp.data]

    # noinspection PyProtocol
    def song_list_hot_comments(self, song: BriefSongProtocol) -> List[BCommentModel]:
        if song.identifier.startswith('audio_'):
            _, id_ = song.identifier.split('_')
            resp = self._api.video_get_hot_comments(VideoHotCommentsRequest(
                type=CommentType.audio, oid=int(id_), ps=10
            ))
        else:
            resp = self._api.video_get_hot_comments(VideoHotCommentsRequest(
                type=CommentType.video, oid=self._get_video_avid(song.identifier), ps=10
            ))
        return [BCommentModel.create_model(comment) for comment in resp.data.replies]

    def song_get_lyric(self, song) -> Optional[LyricModel]:
        if is_child_model(song):
            return None

        if not hasattr(song, 'lyric') or song.lyric is None or len(song.lyric) == 0:
            song = self.song_get(song.identifier)
        if not song.lyric:
            return None
        if song.lyric.endswith('.json'):
            json_data = self._api.get_content(song.lyric)
            return LyricModel(
                source=__identifier__,
                identifier=song.identifier,
                content=json_to_lrc_text(json_data),
            )
        return None

    def _get_video_avid(self, bvid: str) -> int:
        avid = self._video_avids.get(bvid)
        if avid is None:
            info = self._api.video_get_info(VideoInfoRequest(bvid=bvid))
            self._video_avids[bvid] = info.data.aid
            avid = info.data.aid
        return avid

    def _get_video_id_pair(self, identifier, use_first_page=False):
        """
        Some videos/songs has multiple pages, if use_first_page is specified,
        the first page will be used. Otherwise, raise MediaNotFound error
        (because we do not know which page to return).
        """
        if identifier.startswith('paged_'):
            bvid, page_num = parse_paged_identifier(identifier)
            page_num = int(page_num)
        else:
            bvid = identifier
            page_num = -1 if use_first_page else -2

        cid = self._video_cids.get(bvid)
        if cid is None:
            info = self._api.video_get_info(VideoInfoRequest(bvid=bvid))
            if page_num >= 0:
                cid = info.data.pages[page_num - 1].cid
            else:
                if len(info.data.pages) > 1 and page_num != -1:
                    raise MediaNotFound(reason=MediaNotFound.Reason.check_children)
                self._video_cids[bvid] = info.data.cid
                cid = info.data.cid
        return bvid, cid

    def _get_video_dimension(self, bvid):
        info = self._api.video_get_info(VideoInfoRequest(bvid=bvid))
        return info.data.dimension.width, info.data.dimension.height

    def video_list_quality(self, video) -> List[Quality.Video]:
        if video.identifier.startswith('live_'):
            return [Quality.Video.hd]

        bvid, cid = self._get_video_id_pair(video.identifier, use_first_page=True)
        response = self._api.video_get_url(PlayUrlRequest(
            bvid=bvid,
            cid=cid,
            fnval=VideoFnval.FLV
        ))
        qualities = set([q.get_quality() for q in response.data.accept_quality])
        self._video_quality_codes[video.identifier] = response.data.accept_quality
        return list(qualities)

    def song_get_mv(self, song) -> Optional[VideoModel]:
        if song.identifier.startswith('audio_'):
            return None
        if is_child_model(song):
            return None
        song = self.song_get(song.identifier)
        return VideoModel(
            source=__identifier__,
            identifier=song.identifier,
            title=song.title,
            artists=song.artists,
            duration=song.duration,
            cover=''
        )

    def _get_video_danmaku(self, cid, bvid) -> Optional[Path]:
        # find existed ass file
        for p in DANMAKU_DIRECTORY.glob(f'{cid}.ass'):
            if p.is_file() and time() - p.stat().st_mtime < 300:
                return p
        # find existed xml file
        for p in DANMAKU_DIRECTORY.glob(f'{cid}.xml'):
            if p.is_file() and time() - p.stat().st_mtime < 300:
                width, height = self._get_video_dimension(bvid)
                target_file = p.with_suffix('.ass')
                font_size = 50 / 1080 * height
                Danmaku2ASS(p.as_posix(), 'autodetect', target_file.as_posix(), width, height, font_size=font_size)
                return target_file
        # get latest danmuku and transform to ass
        data = self._api.get_content_raw(f'https://comment.bilibili.com/{cid}.xml')
        xml_file = DANMAKU_DIRECTORY / f'{cid}.xml'
        xml_file.unlink(missing_ok=True)
        with xml_file.open('wb') as f:
            f.write(data)
            f.flush()
        target_file = xml_file.with_suffix('.ass')
        width, height = self._get_video_dimension(bvid)
        font_size = 50 / 1080 * height
        Danmaku2ASS(xml_file.as_posix(), 'autodetect', target_file.as_posix(), width, height, font_size=font_size)
        return target_file

    def video_get_media(self, video, quality: Quality.Video) -> Optional[Media]:
        if video.identifier.startswith('live_'):
            _, id_ = video.identifier.split('_')
            resp = self._api.live_play_url(LivePlayUrlRequest(cid=int(id_)))
            if resp.data.durl is None or len(resp.data.durl) == 0:
                return None
            return Media(resp.data.durl[0].url, format='m3u8',
                         http_headers={'Referer': 'https://www.bilibili.com/'})
        bvid, cid = self._get_video_id_pair(video.identifier, use_first_page=True)
        response = self._api.video_get_url(PlayUrlRequest(
            bvid=bvid,
            qn=VideoQualityNum.q1080p60,
            cid=cid,
            fnval=VideoFnval.DASH
        ))
        # select audio
        audios = sorted(response.data.dash.audio, key=lambda a: a.bandwidth, reverse=True)
        if audios is None or len(audios) == 0:
            return None
        # select video
        max_quality_code = VideoQualityNum.get_max_from_quality(quality)
        select_quality = max(filter(lambda c: c.value <= max_quality_code, self._video_quality_codes[video.identifier]),
                             key=lambda c: c.value)
        videos = sorted(filter(lambda a: a.id <= select_quality.value, response.data.dash.video), key=lambda a: a.id,
                        reverse=True)
        if videos is None or len(videos) == 0:
            return None
        if len(signature(VideoAudioManifest).parameters) == 3:
            danmaku_path = self._get_video_danmaku(cid, video.identifier)
            if not danmaku_path.exists():
                danmaku_path = None
            manifest = VideoAudioManifest(videos[0].base_url, audios[0].base_url,
                                          danmaku_path.as_posix() if danmaku_path is not None else None)
        else:
            manifest = VideoAudioManifest(videos[0].base_url, audios[0].base_url)
        return Media(manifest, format='flv', http_headers={'Referer': 'https://www.bilibili.com/'})

    def song_list_quality(self, song) -> List[Quality.Audio]:
        if song.identifier.startswith('audio_'):
            return [Quality.Audio.hq]

        bvid, cid = self._get_video_id_pair(song.identifier)
        response = self._api.video_get_url(PlayUrlRequest(
            bvid=bvid,
            cid=cid,
            fnval=VideoFnval.DASH
        ))
        song_bitrates = [a.bandwidth for a in response.data.dash.audio]
        qualities = []
        for b in song_bitrates:
            if b <= 120000:
                qualities.append(Quality.Audio.lq)
                continue
            if b <= 256000:
                qualities.append(Quality.Audio.sq)
                continue
            qualities.append(Quality.Audio.hq)
        if response.data.dash.flac is not None and response.data.dash.flac['audio'] is not None:
            qualities.append(Quality.Audio.shq)
        print(list(set(qualities)))
        return list(set(qualities))

    def song_get_media(self, song, quality: Quality.Audio) -> Optional[Media]:
        if song.identifier.startswith('audio_'):
            _, id_ = song.identifier.split('_')
            resp = self._api.audio_get_url(AudioGetUrlRequest(sid=int(id_)))
            if len(resp.data.cdns) < 1:
                return None
            return Media(resp.data.cdns[0], type_=MediaType.audio, format='m4a',
                         http_headers={'Referer': 'https://www.bilibili.com/'})

        bvid, cid = self._get_video_id_pair(song.identifier)
        response = self._api.video_get_url(PlayUrlRequest(
            bvid=bvid,
            cid=cid,
            fnval=VideoFnval.DASH
        ))
        if Quality.Audio.shq == quality:
            if response.data.dash.flac is not None and response.data.dash.flac['audio'] is not None:
                return Media(response.data.dash.flac['audio']['base_url'], type_=MediaType.audio, format='m4s',
                             bitrate=int(response.data.dash.flac['audio']['bandwidth'] / 1000),
                             http_headers={'Referer': 'https://www.bilibili.com/'})
        audios = sorted(response.data.dash.audio, key=lambda a: a.bandwidth, reverse=True)
        selects: Optional[List[PlayUrlResponse.PlayUrlResponseData.Dash.DashItem]] = None
        match quality:
            case Quality.Audio.lq:
                selects = list(filter(lambda a: a.bandwidth <= 120000, audios))
            case Quality.Audio.sq:
                selects = list(filter(lambda a: a.bandwidth <= 256000, audios))
            case Quality.Audio.hq:
                selects = audios
        if selects is None or len(selects) == 0:
            return None
        return Media(selects[0].base_url, type_=MediaType.audio, format='m4s', bitrate=int(selects[0].bandwidth / 1000),
                     http_headers={'Referer': 'https://www.bilibili.com/'})

    def user_playlists(self, identifier) -> List[BriefPlaylistModel]:
        resp = self._api.favorite_list(FavoriteListRequest(up_mid=int(identifier)))
        return BPlaylistModel.create_model_list(resp)

    def user_playlist_new(self, title: str, intro: str, privacy: int = None):
        return self._api.favorite_new(FavoriteNewRequest(title=title, intro=intro, privacy=privacy))

    def fav_playlists(self, identifier) -> List[BriefPlaylistModel]:
        resp = self._api.collected_favorite_list(CollectedFavoriteListRequest(up_mid=int(identifier), ps=40))
        return BPlaylistModel.create_model_list(resp)

    def user_following(self) -> List[BriefArtistModel]:
        resp = self._api.user_following(UserFollowingRequest(vmid=int(self._user.identifier), ps=100))
        return [BriefArtistModel(
            source=__identifier__,
            identifier=user.mid,
            name=user.uname,
        ) for user in resp.data.list]

    def current_user_list_playlists(self):
        p1 = self.special_playlists()
        p2 = self.user_playlists(self.get_current_user().identifier)
        p3 = self.audio_favorite_playlists()
        p4 = self.audio_collected_playlists()
        return p1 + p2 + p3 + p4

    def current_user_fav_create_playlists_rd(self):
        p1 = self.fav_playlists(self.get_current_user().identifier)
        p2 = self.audio_collected_playlists()
        return p1 + p2

    def audio_favorite_playlists(self) -> List[BriefPlaylistModel]:
        resp = self._api.audio_favorite_list(PaginatedRequest(ps=100, pn=1))
        return BPlaylistModel.create_audio_model_list(resp)

    def audio_collected_playlists(self) -> List[BriefPlaylistModel]:
        resp = self._api.audio_collected_list(PaginatedRequest(ps=100, pn=1))
        return BPlaylistModel.create_audio_model_list(resp)

    def home_recommend_videos(self, idx) -> List[BriefSongModel]:
        resp = self._api.home_recommend_videos(HomeRecommendVideosRequest(ps=10, fresh_idx=idx, fresh_idx_1h=idx))
        return [BSongModel.create_history_brief_model(v) for v in resp.data.item]

    def most_popular_videos(self) -> List[BriefSongModel]:
        resp = self._api.video_most_popular()
        return [BSongModel.create_history_brief_model(v) for v in resp.data.list]

    def weekly_video_playlists(self) -> List[BriefPlaylistModel]:
        resp = self._api.video_weekly_list()
        return [BPlaylistModel.weekly_brief_model(v) for v in resp.data.list]

    def video_live_feeds(self) -> List[BVideoModel]:
        resp = self._api.live_feed_list(AnotherPaginatedRequest(pagesize=30))
        return [BVideoModel.create_live_model(live) for live in resp.data.list]

    def media_user_collect(self) -> List[BriefAlbumModel]:
        resp1 = self._api.media_bangumi_favlist(MediaFavlistRequest(vmid=int(self._user.identifier), ps=30))
        resp2 = self._api.media_bangumi_favlist(MediaFavlistRequest(vmid=int(self._user.identifier), ps=30, type=2))
        return [BSearchModel.search_media_model(m) for m in resp1.data.list] + \
            [BSearchModel.search_media_model(m) for m in resp2.data.list]

    def audio_playlist_get(self, identifier: str) -> Optional[BPlaylistModel]:
        _, type_, id_ = identifier.split('_')
        match int(type_):
            case 1:
                resp = self._api.audio_favorite_info(AudioFavoriteSongsRequest(sid=int(id_)))
                return BPlaylistModel.create_audio_model(resp.data)
            case 2:
                resp = self._api.audio_collected_info(AudioFavoriteSongsRequest(sid=int(id_)))
                return BPlaylistModel.create_audio_model(resp.data)
        return None

    def album_get(self, identifier: str) -> Optional[BAlbumModel]:
        if identifier.startswith('media_'):
            _, __, ssid = identifier.split('_')
            resp = self._api.media_bangumi_get_list(MediaGetListRequest(season_id=ssid))
            return BAlbumModel.create_season_model(resp.result)
        raise ModelNotFound(f'album:{identifier} not found')

    def playlist_get(self, identifier: str) -> Optional[BPlaylistModel]:
        # fixme: fuo should support playlist_get v2 first
        if identifier.startswith('weekly_'):
            return BPlaylistModel(
                source=__identifier__,
                identifier=identifier,
                creator=None,
                name='',
                cover='',
                description='',
                count=0,
            )
        if identifier.startswith('audio_'):
            return self.audio_playlist_get(identifier)
        if identifier == 'LATER':
            resp = self._api.history_later_videos()
            return BPlaylistModel.special_model(identifier, resp)
        if identifier in ['HISTORY', 'DYNAMIC']:
            return BPlaylistModel.special_model(identifier, None)
        fav_type, id_ = identifier.split('_')
        if int(fav_type) == 21:
            resp = self._api.favorite_season_resource(FavoriteSeasonResourceRequest(season_id=int(id_), ps=0))
        else:
            resp = self._api.favorite_info(FavoriteInfoRequest(media_id=int(id_)))
        return BPlaylistModel.create_info_model(resp)

    def audio_playlist_create_songs_rd(self, playlist):
        _, type_, id_ = playlist.identifier.split('_')

        if int(type_) == 2 and playlist.count == 0:
            # 兼容歌单信息不存在歌曲数量的问题
            response = self._api.audio_collected_songs(AudioFavoriteSongsRequest(
                sid=int(id_), pn=1, ps=0
            ))
            playlist.count = response.data.totalSize

        def g():
            page = 1
            while page <= math.ceil(playlist.count / 20):
                match int(type_):
                    case 1:
                        response = self._api.audio_favorite_songs(AudioFavoriteSongsRequest(
                            sid=int(id_), pn=page
                        ))
                        for au in response.data.data:
                            yield BSongModel.create_audio_model(au)
                    case 2:
                        response = self._api.audio_collected_songs(AudioFavoriteSongsRequest(
                            sid=int(id_), pn=page
                        ))
                        for au in response.data.data:
                            yield BSongModel.create_audio_model(au)
                page += 1

        return SequentialReader(g(), playlist.count)

    def weekly_playlist_create_songs_rd(self, playlist):
        _, id_ = playlist.identifier.split('_')

        response = self._api.video_weekly_detail(WeeklyDetailRequest(number=id_))

        def g():
            page = 1
            if page == 1:
                for i in response.data.list:
                    yield BSongModel.create_history_brief_model(i)
            page += 1

        return SequentialReader(g(), len(response.data.list))

    def playlist_create_songs_rd(self, playlist):
        if playlist.identifier.startswith('audio_'):
            return self.audio_playlist_create_songs_rd(playlist)

        if playlist.identifier.startswith('weekly_'):
            return self.weekly_playlist_create_songs_rd(playlist)

        def g():
            _dynamic_offset = None

            if playlist.identifier == 'LATER':
                response = self._api.history_later_videos()
                song_list = BSongModel.create_history_brief_model_list(response)
                for s in song_list:
                    yield s
            else:
                is_season = False
                id_ = None
                if playlist.identifier not in ['HISTORY', 'DYNAMIC']:
                    fav_type, id_ = playlist.identifier.split('_')
                    is_season = int(fav_type) == 21
                page = 1
                while page <= math.ceil(playlist.count / 20):
                    if playlist.identifier == 'DYNAMIC':
                        resp = self._api.home_dynamic_videos(HomeDynamicVideoRequest(offset=_dynamic_offset, page=page))
                        _dynamic_offset = resp.data.offset
                        for v in resp.data.items:
                            yield BSongModel.create_dynamic_brief_model(v)
                        if not resp.data.has_more:
                            return
                    elif playlist.identifier == 'HISTORY':
                        response = self._api.history_videos(PaginatedRequest(pn=page))
                        for m in response.data:
                            yield BSongModel.create_history_brief_model(m)
                    else:
                        if is_season:
                            response = self._api.favorite_season_resource(FavoriteSeasonResourceRequest(
                                season_id=int(id_),
                                pn=page,
                            ))
                        else:
                            response = self._api.favorite_resource(FavoriteResourceRequest(
                                media_id=int(id_),
                                pn=page,
                            ))
                        for m in response.data.medias:
                            yield BSongModel.create_brief_model(m)
                    page += 1

        return SequentialReader(g(), playlist.count)

    @staticmethod
    def special_playlists() -> List[BriefPlaylistModel]:
        return BPlaylistModel.special_brief_playlists()

    def artist_get(self, identifier) -> BArtistModel:
        resp = self._api.user_info(UserInfoRequest(mid=identifier))
        video_resp = self._api.user_best_videos(UserBestVideoRequest(vmid=identifier))
        return BArtistModel.create_model(resp, video_resp)

    def artist_create_songs_rd(self, artist):
        resp = self._api.user_videos(UserVideoRequest(mid=artist.identifier, ps=1, pn=1))
        total = resp.data.page.count

        def g():
            page = 1
            while page <= math.ceil(total / 20):
                response = self._api.user_videos(UserVideoRequest(mid=artist.identifier, ps=20, pn=page))
                for m in response.data.list.vlist:
                    yield BSongModel.create_user_brief_model(m)
                page += 1

        return SequentialReader(g(), total)

    def rec_a_collection_of_videos(self):
        if self.has_current_user():
            resp = self._api.home_recommend_videos(HomeRecommendVideosRequest(ps=12))
            videos = [BVideoModel.create_recommend_video_model(v) for v in resp.data.item]
        else:
            videos = []
        return Collection(
            name='推荐视频',
            type_=CollectionType.only_videos,
            models=videos,
        )

    @property
    def identifier(self):
        return __identifier__

    @property
    def name(self):
        return __alias__

    def close(self):
        self._api.close()

    def __del__(self):
        self.close()


if __name__ == '__main__':
    from inspect import signature

    print(len(signature(VideoAudioManifest).parameters))



from typing import Type, Any, Optional

from fuo_bilibili.api.schema.requests import VideoInfoRequest, PlayUrlRequest, BaseRequest, VideoHotCommentsRequest, \
    WeeklyDetailRequest
from fuo_bilibili.api.schema.responses import VideoInfoResponse, PlayUrlResponse, BaseResponse, VideoGetRelatedResponse, \
    VideoHotCommentsResponse, VideoMostPopularResponse, VideoWeeklyListResponse, WeeklyDetailResponse


class VideoMixin:
    API_BASE = 'https://api.bilibili.com/x/web-interface'
    PLAYER_API_BASE = 'https://api.bilibili.com/x/player'
    APIX_BASE = 'https://api.bilibili.com/x'

    def get(self, url: str, param: Optional[BaseRequest], clazz: Type[BaseResponse]) -> Any:
        pass

    def video_get_info(self, request: VideoInfoRequest) -> VideoInfoResponse:
        url = f'{self.API_BASE}/view'
        return self.get(url, request, VideoInfoResponse)

    def video_get_url(self, request: PlayUrlRequest) -> PlayUrlResponse:
        url = f'{self.PLAYER_API_BASE}/playurl'
        return self.get(url, request, PlayUrlResponse)

    def video_get_related(self, request: VideoInfoRequest) -> VideoGetRelatedResponse:
        url = f'{self.API_BASE}/archive/related'
        return self.get(url, request, VideoGetRelatedResponse)

    def video_get_hot_comments(self, request: VideoHotCommentsRequest) -> VideoHotCommentsResponse:
        url = f'{self.APIX_BASE}/v2/reply'
        return self.get(url, request, VideoHotCommentsResponse)

    def video_most_popular(self) -> VideoMostPopularResponse:
        url = f'{self.APIX_BASE}/web-interface/popular'
        return self.get(url, None, VideoMostPopularResponse)

    def video_weekly_list(self) -> VideoWeeklyListResponse:
        url = f'{self.APIX_BASE}/web-interface/popular/series/list'
        return self.get(url, None, VideoWeeklyListResponse)

    def video_weekly_detail(self, request: WeeklyDetailRequest) -> WeeklyDetailResponse:
        url = f'{self.APIX_BASE}/web-interface/popular/series/one'
        return self.get(url, request, WeeklyDetailResponse)



from typing import Type, Any, Optional, Union

from .compat import BaseModel

from fuo_bilibili.api.schema.requests import BaseRequest, UserInfoRequest, UserBestVideoRequest, UserVideoRequest, \
    UserFollowingRequest
from fuo_bilibili.api.schema.responses import BaseResponse, UserInfoResponse, UserBestVideoResponse, UserVideoResponse, \
    UserFollowingResponse


class UserMixin:
    API_BASE = 'https://api.bilibili.com/x/web-interface'
    PASSPORT_BASE = 'https://passport.bilibili.com'
    APIX_BASE = 'https://api.bilibili.com/x'

    def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) -> Any:
        pass

    def get_uncached(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) -> Any:
        pass

    def user_info(self, param: UserInfoRequest) -> UserInfoResponse:
        url = f'{self.APIX_BASE}/space/wbi/acc/info'
        return self.get_uncached(url, param, UserInfoResponse, headers={
            'user-agent': 'Mozilla/5.0',
            'referer': 'https://www.bilibili.com',
        })

    def user_best_videos(self, request: UserBestVideoRequest) -> UserBestVideoResponse:
        url = f'{self.APIX_BASE}/space/masterpiece'
        return self.get(url, request, UserBestVideoResponse)

    def user_videos(self, request: UserVideoRequest) -> UserVideoResponse:
        url = f'{self.APIX_BASE}/space/wbi/arc/search'
        return self.get(url, request, UserVideoResponse)

    def user_following(self, request: UserFollowingRequest) -> UserFollowingResponse:
        url = f'{self.APIX_BASE}/relation/followings'
        return self.get(url, request, UserFollowingResponse)



from typing import Optional, Union

from feeluown.app import App

__alias__ = '哔哩哔哩'
__feeluown_version__ = '3.5'
__version__ = '0.0.1'
__desc__ = __alias__
__identifier__ = 'bilibili'

from fuo_bilibili.provider import BilibiliProvider

provider = BilibiliProvider()

def init_config(config):
    config.deffield('ENABLE_LIVE_ROOM_AS_VIDEO',
                    type_=bool,
                    default=True,
                    desc='treat live room as video')


# noinspection PyProtectedMember
def enable(app):
    provider.enable_live_room_as_video = app.config.bilibili.ENABLE_LIVE_ROOM_AS_VIDEO
    app.library.register(provider)
    if app.mode & App.GuiMode:
        from fuo_bilibili.ui import BProviderUi
        ui_mgr = BProviderUi(app, provider)
        app.pvd_ui_mgr.register(ui_mgr)


def disable(app: App):
    app.library.deregister(provider)
    if app.mode & App.GuiMode:
        provider.close()
        # noinspection PyUnresolvedReferences
        app.providers.remove(provider.identifier)
        app.pvd_ui_mgr.remove(provider.identifier)



from typing import Type, Any, Optional, Union

from .compat import BaseModel

from fuo_bilibili.api.schema.requests import BaseRequest, AnotherPaginatedRequest, LivePlayUrlRequest
from fuo_bilibili.api.schema.responses import BaseResponse, LiveFeedListResponse, LivePlayUrlResponse


class LiveMixin:
    API_LIVE_BASE = 'https://api.live.bilibili.com'

    def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None]) -> Any:
        pass

    def get_uncached(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) -> Any:
        pass

    def post(self, url: str, param: Optional[BaseRequest], clazz: Type[BaseResponse], is_json=False, **kwargs) -> Any:
        pass

    def live_feed_list(self, request: AnotherPaginatedRequest) -> LiveFeedListResponse:
        url = f'{self.API_LIVE_BASE}/relation/v1/feed/feed_list'
        return self.get(url, request, LiveFeedListResponse)

    def live_play_url(self, request: LivePlayUrlRequest) -> LivePlayUrlResponse:
        url = f'{self.API_LIVE_BASE}/room/v1/Room/playUrl'
        return self.get(url, request, LivePlayUrlResponse)



from datetime import datetime, timedelta
from typing import Any, List, Optional, Union

from fuo_bilibili.api.compat import BaseModel, Field, field_validator

from fuo_bilibili.api.schema.enums import VideoQualityNum, CodecId, VideoCopyright, VideoState, VideoFrom, SearchType, \
    VipType, MediaType


class BaseResponse(BaseModel):
    code: int = None
    message: str = None
    ttl: int = None
    data: Any


class Dimension(BaseModel):
    width: int
    height: int
    rotate: int


class Upper(BaseModel):
    mid: int
    name: str
    face: str = ''
    followed: bool = None
    vip_type: VipType = None
    vip_statue: bool = None


class CntInfo(BaseModel):
    collect: int
    play: int
    danmaku: int = None


class Owner(BaseModel):
    mid: int  # UID
    name: str  # 用户名
    face: str  # 头像


class Stat(BaseModel):
    aid: int = None  # AV号
    view: int  # 播放量
    danmaku: int  # 弹幕量
    reply: int = None  # 评论量
    favorite: int = None  # 收藏量
    coin: int = None  # 投币量
    share: int = None  # 分享量
    now_rank: int = None  # 当前排名
    his_rank: int = None  # 历史最高排名
    like: int  # 点赞量
    dislike: int = None  # 点踩量
    evaluation: str = None  # 视频评分
    argue_msg: str = None  # 警告提示信息


class Rights(BaseModel):
    bp: int
    elec: bool  # 支持充电
    download: bool  # 允许下载
    movie: bool  # 是否是电影
    pay: bool  # 是否PGC付费
    hd5: bool  # 是否支持高码率
    no_reprint: bool  # 是否显示禁止转载
    autoplay: bool  # 是否自动播放
    ugc_pay: bool  # 是否UGC付费
    is_cooperation: bool  # 是否联合投稿
    ugc_pay_preview: bool  # UGC付费预览
    no_background: bool
    clean_mode: bool = None
    is_stein_gate: bool = None
    is_360: bool = None
    no_share: bool = None
    arc_pay: bool
    free_watch: bool = None


class SearchResultKetang(BaseModel):
    type: SearchType = SearchType.KETANG


class SearchResultUser(BaseModel):
    type: SearchType = SearchType.BILI_USER
    mid: int
    uname: str
    usign: str
    fans: int
    videos: int
    upic: str


class SearchResultMedia(BaseModel):
    class MediaScore(BaseModel):
        user_count: int
        score: float

    type: SearchType = SearchType.MEDIA
    media_id: int  # mdid
    season_id: int  # ssid
    title: str
    org_title: str
    cover: str
    media_type: MediaType
    areas: str
    styles: str
    cv: str
    staff: str
    desc: str
    corner: int
    media_score: MediaScore = None
    season_type_name: str
    is_follow: bool


class SearchResultLiveRoom(BaseModel):
    type: SearchType = SearchType.LIVE_ROOM
    uid: int
    tags: str  # 直播间tag
    live_time: datetime
    cate_name: str  # 子分区名
    live_status: bool
    uname: str
    uface: str
    user_cover: str  # 封面
    short_id: str  # 直播间短号
    title: str  # 标题
    cover: str  # 直播关键帧
    online: int  # 在线人数
    roomid: int  # 直播间id
    attentions: int  # 粉丝数


class SearchResultVideo(BaseModel):
    type: SearchType = SearchType.VIDEO
    id: int  # av号
    author: str  # UPer
    mid: int  # UID
    typeid: int  # 分区id
    typename: str  # 分区名称
    arcurl: str
    aid: int
    bvid: str  # bv号
    title: str  # 标题
    description: str  # 简介
    pic: str  # 视频封面
    play: int  # 播放量
    video_review: int  # 弹幕量
    favorites: int  # 收藏量
    tag: str  # 视频标签
    review: int  # 评论数
    pubdate: datetime
    senddate: datetime
    duration: timedelta
    badgepay: bool
    hit_columns: List[str]
    view_type: str
    is_pay: bool
    is_union_video: bool
    # 从某个时间点开始，这两个字段没了
    # arcrank: str
    # rank_score: int


class RequestCaptchaResponse(BaseResponse):
    class RequestCaptchaResponseData(BaseModel):
        class Geetest(BaseModel):
            gt: str
            challenge: str

        geetest: Geetest
        tencent: dict
        token: str
        type: str

    data: RequestCaptchaResponseData = None


class RequestLoginKeyResponse(BaseModel):
    hash: str
    key: str


class CookieInfoResponse(BaseResponse):
    class CookieInfoResponseData(BaseModel):
        refresh: bool
        timestamp: int

    data: CookieInfoResponseData = None


class SendSmsCodeResponse(BaseResponse):
    class SendSmsCodeResponseData(BaseModel):
        captcha_key: str

    data: SendSmsCodeResponseData = None


class PasswordLoginResponse(BaseResponse):
    class PasswordLoginResponseData(BaseModel):
        status: int = None
        message: str = None
        url: str = None

    data: PasswordLoginResponseData = None


class SmsCodeLoginResponse(BaseResponse):
    class SmsCodeLoginResponseData(BaseModel):
        url: str

    data: SmsCodeLoginResponseData = None


class NavInfoResponse(BaseResponse):
    class NavInfoResponseData(BaseModel):
        class LevelInfo(BaseModel):
            current_level: int
            current_min: int
            current_exp: int
            next_exp: str

        class VipLabel(BaseModel):
            text: str

        class Wallet(BaseModel):
            bcoin_balance: int  # B币
            coupon_balance: int  # 赠送B币数

        class Wbi(BaseModel):
            img_url: str
            sub_url: str

        isLogin: bool
        email_verified: bool = None
        face: str = None
        mid: int = None  # UID
        mobile_verified: bool = None
        money: float = None  # 硬币数
        moral: int = None  # 节操值
        uname: str = None
        vipDueDate: datetime = None
        vipStatus: bool = None
        vipType: VipType = None
        vip_pay_type: bool = None
        wallet: Wallet = None
        level_info: LevelInfo = None
        vip_label: VipLabel = None
        wbi_img: Wbi = None

        @field_validator('vipDueDate', mode='before')
        @classmethod
        def convert_ms(cls, v):
            return v / 1000

    data: NavInfoResponseData = None


class SearchResponse(BaseResponse):
    class SearchResponseData(BaseModel):
        seid: int
        page: int
        pagesize: int  # 每页条数
        numResults: int  # 总数
        numPages: int  # 总页数
        suggest_keyword: str
        rqt_type: str
        # 从某个时间开始，服务端不返回这个字段了
        # cost_time: dict
        exp_list: dict = None
        egg_hit: int
        pageinfo: dict = None
        result: Union[List[Union[SearchResultVideo, SearchResultUser,
        SearchResultLiveRoom, SearchResultMedia,
        SearchResultKetang]], dict] = None
        show_column: int

    data: SearchResponseData = None


class VideoInfoResponse(BaseResponse):
    class VideoInfoResponseData(BaseModel):
        class DescV2(BaseModel):
            raw_text: str
            type: int
            biz_id: int = None

        class Page(BaseModel):
            cid: int  # 分P cid
            page: int  # 当前分P
            from_: VideoFrom = Field(alias='from')  # 视频来源
            part: str  # 分P标题
            duration: timedelta
            vid: str = None  # 站外视频id
            weblink: str = None  # 站外视频链接
            dimension: Dimension

        class Subtitle(BaseModel):
            class SubtitleListItem(BaseModel):
                class Author(BaseModel):
                    mid: int  # 上传者UID
                    name: str  # 昵称
                    sex: str  # 性别
                    face: str  # 头像
                    sign: str  # 签名
                    rank: int
                    birthday: int
                    is_fake_account: bool
                    is_deleted: bool

                id: int
                lan: str  # 语言
                lan_doc: str  # 语言名
                is_lock: bool  # 是否锁定
                author_mid: int = None  # 上传者UID
                subtitle_url: str  # 字幕文件 json
                author: Author

            allow_submit: bool  # 是否允许提交字幕
            list: List[SubtitleListItem]

        class Staff(BaseModel):
            class Vip(BaseModel):
                type: int  # 会员类型 0：无 1：月会员 2：年会员
                status: bool
                theme_type: int

            mid: int  # UID
            title: str  # 成员名称
            name: str  # 成员昵称
            face: str  # 头像
            vip: Vip
            official: dict
            follower: int  # 粉丝数

        bvid: str
        aid: int
        videos: int  # 分P总数
        tid: int  # 分区id
        tname: str  # 子分区名称
        copyright: VideoCopyright
        pic: str  # 封面
        title: str  # 标题
        pubdate: datetime  # 稿件发布时间
        ctime: datetime  # 用户投稿时间
        desc: str  # 简介
        desc_v2: List[DescV2] = None  # 新版视频简介
        state: VideoState
        duration: timedelta
        forward: int = None
        mission_id: int = None
        redirect_url: str = None
        rights: Rights
        owner: Owner
        stat: Stat
        dynamic: str
        cid: int  # 1P cid
        dimension: Dimension
        no_cache: bool
        pages: List[Page]
        subtitle: Subtitle
        staff: List[Staff] = None
        user_garb: dict

    data: VideoInfoResponseData = None


class PlayUrlResponse(BaseResponse):
    class PlayUrlResponseData(BaseModel):
        class Durl(BaseModel):
            order: int
            length: timedelta  # 视频长度秒
            size: int  # 视频大小 Bytes
            url: str  # 视频流url
            backup_url: List[str] = None  # 备用视频流

            @field_validator('length', mode='before')
            @classmethod
            def convert_ms(cls, v):
                return v / 1000

        class Dash(BaseModel):
            class DashItem(BaseModel):
                id: int  # 音视频清晰度代码
                base_url: str  # 默认视频/音频流
                backup_url: List[str] = None  # 备用视频/音频流
                bandwidth: int
                mime_type: str
                codecs: str
                width: int
                height: int
                frame_rate: str  # 视频帧率
                sar: str  # Sample Aspect Ratio
                start_with_sap: int  # Stream Access Point
                segment_base: dict
                codecid: CodecId

            duration: timedelta  # 视频长度秒
            video: List[DashItem]
            audio: List[DashItem]
            flac: Optional[dict] = None
            dolby: Optional[dict] = None

        quality: VideoQualityNum
        format: str
        timelength: timedelta  # 视频长度毫秒
        accept_format: str  # 支持的格式
        accept_description: List[str]  # 支持的分辨率列表描述
        accept_quality: List[VideoQualityNum]  # 支持的分辨率代码列表
        durl: List[Durl] = None  # vlc/mp4 视频分段
        dash: Dash = None

        @field_validator('timelength', mode='before')
        @classmethod
        def convert_ms(cls, v):
            return v / 1000

    data: PlayUrlResponseData = None


class FavoriteListResponse(BaseResponse):
    class FavoriteListResponseData(BaseModel):
        class FavoriteList(BaseModel):
            id: int  # 收藏id
            fid: int  # 原始收藏夹id
            mid: int  # 创建者UID
            attr: int
            title: str  # 收藏夹标题
            fav_state: bool
            media_count: int  # 收藏夹内容数量
            type: int = 11

        count: int  # 总数
        list: List[FavoriteList]  # 收藏列表

    data: FavoriteListResponseData = None


class CollectedFavoriteListResponse(BaseResponse):
    class CollectedFavoriteListResponseData(BaseModel):
        class CollectedFavoriteList(BaseModel):
            id: int  # 收藏id
            fid: int  # 原始收藏夹id
            mid: int  # 创建者UID
            attr: int
            title: str  # 收藏夹标题
            fav_state: bool
            media_count: int  # 收藏夹内容数量
            cover: str  # 封面
            cover_type: int
            intro: str  # 简介
            link: str
            mtime: datetime
            state: int
            type: int  # 11: 分P 21: 合集
            upper: Upper
            view_count: int  # 总播放

        count: int  # 总数
        list: List[CollectedFavoriteList]  # 收藏列表

    data: CollectedFavoriteListResponseData = None


class FavoriteInfoResponse(BaseResponse):
    class FavoriteInfoResponseData(BaseModel):
        id: int
        fid: int
        mid: int
        attr: int
        title: str  # 标题
        cover: str  # 封面
        upper: Upper  # 创建者信息
        type: int  # 11: 分P 21: 合集
        intro: str  # 备注
        fav_state: bool  # 是否收藏
        like_state: bool  # 是否点赞
        media_count: int  # 收藏夹的内容数量

    data: FavoriteInfoResponseData = None


class FavoriteResourceResponse(BaseResponse):
    class FavoriteResourceResponseData(BaseModel):
        class Media(BaseModel):
            id: int
            type: int
            title: str
            cover: str
            intro: str
            page: int
            duration: timedelta
            upper: Upper
            cnt_info: CntInfo
            link: str  # 视频 deeplink
            # ctime: datetime
            # pubtime: datetime
            # favtime: datetime
            bvid: str

        medias: List[Media] = None

    data: FavoriteResourceResponseData = None


class FavoriteSeasonResourceResponse(BaseResponse):
    class FavoriteSeasonResourceResponseData(BaseModel):
        class Info(BaseModel):
            cnt_info: CntInfo
            cover: str
            id: int
            media_count: int
            season_type: int
            title: str
            upper: Upper

        class Media(BaseModel):
            id: int
            title: str
            cover: str
            duration: timedelta
            upper: Upper
            cnt_info: CntInfo
            # ctime: datetime
            # pubtime: datetime
            # favtime: datetime
            bvid: str

        info: Info
        medias: List[Media] = None

    data: FavoriteSeasonResourceResponseData = None


class HistoryLaterVideoResponse(BaseResponse):
    class HistoryLaterVideoResponseData(BaseModel):
        class LaterItem(BaseModel):
            aid: int
            videos: int  # 分P总数
            tid: int  # 分区
            tname: str  # 分区名
            copyright: VideoCopyright
            pic: str  # 封面
            title: str  # 标题
            desc: str  # 简介
            state: VideoState
            duration: timedelta
            rights: Rights
            owner: Owner
            stat: Stat
            dynamic: str  # 同步动态内容
            dimension: Dimension
            count: int = None  # 分P数
            cid: int  # cid
            progress: timedelta  # 观看进度
            bvid: str

        count: int
        list: List[LaterItem] = None

    data: HistoryLaterVideoResponseData = None


class HistoryVideoResponse(BaseResponse):
    class HistoryVideoResponseData(BaseModel):
        aid: int
        videos: int  # 分P总数
        tid: int  # 分区
        tname: str  # 分区名
        copyright: VideoCopyright
        pic: str  # 封面
        title: str  # 标题
        desc: str  # 简介
        state: VideoState
        duration: timedelta
        rights: Rights
        owner: Owner
        stat: Stat
        dynamic: str  # 同步动态内容
        dimension: Dimension
        cid: int  # cid
        favorite: bool
        type: int
        sub_type: int
        device: int
        count: int = None  # 分P数
        progress: timedelta  # 观看进度
        bvid: str

    data: List[HistoryVideoResponseData] = None


class HomeRecommendVideosResponse(BaseResponse):
    class HomeRecommendVideosResponseData(BaseModel):
        class Video(BaseModel):
            bvid: str
            cid: int
            duration: timedelta
            id: int
            is_followed: bool
            owner: Owner
            pic: str
            rcmd_reason: dict = None
            stat: Stat
            title: str
            uri: str

        item: List[Video]

    data: HomeRecommendVideosResponseData = None


class HomeDynamicVideoResponse(BaseResponse):
    class HomeDynamicVideoResponseData(BaseModel):
        class DynamicVideoItem(BaseModel):
            class Modules(BaseModel):
                class ModuleDynamic(BaseModel):
                    class Major(BaseModel):
                        class Archive(BaseModel):
                            aid: int
                            bvid: str
                            cover: str
                            desc: str
                            duration_text: str
                            title: str

                        archive: Archive

                    major: Major

                class ModuleAuthor(BaseModel):
                    mid: int
                    name: str
                    face: str

                module_dynamic: ModuleDynamic
                module_author: ModuleAuthor

            modules: Modules

        has_more: bool
        offset: str
        items: List[DynamicVideoItem]

    data: HomeDynamicVideoResponseData = None


class UserInfoResponse(BaseResponse):
    class UserInfoResponseData(BaseModel):
        class FansMedal(BaseModel):
            class Medal(BaseModel):
                target_id: int  # 勋章对应UP的UID
                medal_name: str  # 粉丝勋章名称
                level: int  # 等级
                medal_color: int

            show: bool
            wear: bool
            medal: Medal = None

        mid: int
        name: str
        sex: str
        face: str
        face_nft: bool
        sign: str
        rank: int = None
        level: int
        silence: bool
        fans_badge: bool
        fans_medal: FansMedal
        is_followed: bool
        top_photo: str  # 头图

    data: UserInfoResponseData = None


class UserBestVideoResponse(BaseResponse):
    class BestVideo(BaseModel):
        aid: int
        videos: int
        tid: int
        tname: str
        copyright: VideoCopyright
        pic: str
        title: str
        desc: str
        state: VideoState
        duration: timedelta
        owner: Owner
        dynamic: str
        stat: Stat
        cid: int
        bvid: str
        dimension: Dimension
        reason: str
        inter_video: bool

    data: List[BestVideo]


class UserVideoResponse(BaseResponse):
    class UserVideoResponseData(BaseModel):
        class RList(BaseModel):
            class Video(BaseModel):
                aid: int
                bvid: str
                author: str
                description: str
                length: str  # MM:SS
                mid: int
                title: str
                pic: str

            tlist: dict
            vlist: List[Video]

        class Page(BaseModel):
            count: int

        list: RList
        page: Page

    data: UserVideoResponseData = None


class AudioPlaylist(BaseModel):
    id: int = None  # 音频收藏夹mlid
    uid: int
    uname: str
    title: str
    type: int
    published: bool = None  # 是否公开
    cover: str
    song: int = None  # 音乐数量
    snum: int = None  # 音乐数量
    desc: str = None
    intro: str = None
    sids: List[int] = None  # 音频id列表
    menuId: int  # 对应歌单id
    statistic: dict


class AudioFavoriteListResponse(BaseResponse):
    class AudioFavoriteListResponseData(BaseModel):
        curPage: int
        pageCount: int
        totalSize: int
        pageSize: int
        data: List[AudioPlaylist]

    data: AudioFavoriteListResponseData = None


class AudioPlaylistSong(BaseModel):
    aid: int
    author: str
    bvid: str
    cid: int
    cover: str
    duration: timedelta
    id: int
    intro: str
    lyric: str
    title: str
    uid: int
    uname: str


class AudioFavoriteSongsResponse(BaseResponse):
    class AudioFavoriteSongsResponseData(BaseModel):
        curPage: int
        pageCount: int
        totalSize: int
        pageSize: int
        data: List[AudioPlaylistSong]

    data: AudioFavoriteSongsResponseData = None


class AudioFavoriteInfoResponse(BaseResponse):
    data: AudioPlaylist = None


class AudioGetUrlResponse(BaseResponse):
    class AudioGetUrlResponseData(BaseModel):
        cdns: List[str]
        sid: int
        size: int
        type: int

    data: AudioGetUrlResponseData = None


class VideoGetRelatedResponse(BaseResponse):
    class VideoGetRelatedResponseData(BaseModel):
        bvid: str
        aid: int
        videos: int  # 分P总数
        tid: int  # 分区id
        tname: str  # 子分区名称
        copyright: VideoCopyright
        pic: str  # 封面
        title: str  # 标题
        desc: str  # 简介
        state: VideoState
        duration: timedelta
        forward: int = None
        mission_id: int = None
        redirect_url: str = None
        rights: Rights
        owner: Owner
        stat: Stat
        dynamic: str
        cid: int  # 1P cid
        dimension: Dimension

    data: List[VideoGetRelatedResponseData] = None


class VideoHotCommentsResponse(BaseResponse):
    class VideoHotCommentsResponseData(BaseModel):
        class Reply(BaseModel):
            class Member(BaseModel):
                mid: int
                uname: str
                avatar: str

            class Content(BaseModel):
                message: str
                # plat: int
                # device: str

            rpid: int
            like: int
            ctime: datetime
            member: Member
            content: Content

        page: dict
        replies: List[Reply]

    data: VideoHotCommentsResponseData = None


class LiveFeedListResponse(BaseResponse):
    class LiveFeedListResponseData(BaseModel):
        class LiveFeed(BaseModel):
            area_id: int
            cover: str
            face: str
            link: str
            online: int
            parent_area_id: int
            pic: str
            roomid: int
            title: str
            uid: int
            uname: str

        page: int
        pagesize: int
        results: int
        list: List[LiveFeed]

    data: LiveFeedListResponseData = None


class LivePlayUrlResponse(BaseResponse):
    class LivePlayUrlResponseData(BaseModel):
        class Qn(BaseModel):
            qn: int
            desc: str

        class Durl(BaseModel):
            url: str
            order: int

        current_quality: int
        accept_quality: List[int]
        current_qn: int
        quality_description: List[Qn]
        durl: List[Durl]

    data: LivePlayUrlResponseData = None


class MediaGetListResponse(BaseResponse):
    class MediaGetListResponseData(BaseModel):
        class Episode(BaseModel):
            aid: int
            bvid: str
            badge: str  # 标签
            cid: int
            cover: str
            title: str
            subtitle: str
            vid: str
            long_title: str
            duration: timedelta

            @field_validator('duration', mode='before')
            @classmethod
            def convert_ms(cls, v):
                return v / 1000

        bkg_cover: str = None
        cover: str
        episodes: List[Episode]
        evaluate: str  # 简介
        media_id: int
        season_id: int
        seasons: list = None
        section: list = None
        share_sub_title: str  # 备注
        square_cover: str  # 方形封面
        title: str
        subtitle: str
        type: MediaType
        up_info: dict = None

    result: MediaGetListResponseData = None
    data: Any = None


class MediaFavlistResponse(BaseResponse):
    class MediaFavlistResponseData(BaseModel):
        class MediaInfo(BaseModel):
            media_id: int
            season_id: int
            season_type: int
            season_type_name: str
            title: str
            cover: str
            total_count: int  # 总集数 -1 未完结
            is_finish: bool  # 是否完结
            evaluate: str  # 简介
            subtitle: str

        pn: int
        ps: int
        total: int
        list: List[MediaInfo]

    data: MediaFavlistResponseData = None


class UserFollowingResponse(BaseResponse):
    class UserFollowingResponseData(BaseModel):
        class UserFollowing(BaseModel):
            mid: int
            uname: str
            face: str
            sign: str

        total: int
        list: List[UserFollowing]

    data: UserFollowingResponseData = None


class VideoMostPopularResponse(BaseResponse):
    class VideoMostPopularResponseData(BaseModel):
        class PopularVideo(BaseModel):
            aid: int
            bvid: str
            videos: int
            tid: int
            tname: str
            copyright: VideoCopyright
            pic: str
            title: str
            pubdate: datetime
            ctime: datetime
            desc: str
            duration: timedelta
            owner: Owner
            stat: Stat
            rcmd_reason: dict

        list: List[PopularVideo]

    data: VideoMostPopularResponseData = None


class VideoWeeklyListResponse(BaseResponse):
    class VideoWeeklyListResponseData(BaseModel):
        class VideoWeeklyItem(BaseModel):
            number: int
            subject: str
            status: int
            name: str

        list: List[VideoWeeklyItem]

    data: VideoWeeklyListResponseData = None


class WeeklyDetailResponse(BaseResponse):
    class WeeklyDetailResponseData(BaseModel):
        class WeeklyDetail(BaseModel):
            aid: int
            videos: int
            tid: int
            tname: str
            copyright: VideoCopyright
            pic: str
            title: str
            duration: timedelta
            rights: Rights
            owner: Owner
            stat: Stat
            dynamic: str
            cid: int
            dimension: Dimension
            short_link: str
            short_link_v2: str
            first_frame: str
            pub_location: str
            bvid: str
            rcmd_reason: str

        list: List[WeeklyDetail]

    data: WeeklyDetailResponseData = None



import json
from datetime import timedelta
from typing import Optional


def rsa_encrypt(text: str, public_key: str) -> str:
    """
    RSA公钥加密
    :param text:
    :param public_key:
    :return:
    """
    import base64
    from Crypto.PublicKey import RSA
    from Crypto.Cipher import PKCS1_v1_5
    key = RSA.import_key(public_key.encode())
    encrypted = PKCS1_v1_5.new(key).encrypt(text.encode())
    return base64.b64encode(encrypted).decode()


def format_timedelta_to_hms(td: timedelta) -> str:
    ss = int(td.total_seconds())
    mm = int(ss / 60)
    ss -= mm * 60
    hh = 0
    if mm >= 60:
        hh = int(mm / 60)
        mm -= hh * 60
    result = [str(mm).zfill(2), str(ss).zfill(2)]
    if hh > 0:
        result.insert(0, str(hh).zfill(2))
    return ':'.join(result)


def json_to_lrc_text(jsons: str) -> Optional[str]:
    if jsons is None or len(jsons) == 0:
        return None
    data = json.loads(jsons)
    body = data.get('body')
    if body is None or len(body) == 0:
        return None
    lrc_lines = []
    for item in body:
        from_second = int(item['from'])
        from_ms = str(item['from'] - from_second).lstrip('0')
        from_second_str = format_timedelta_to_hms(timedelta(seconds=from_second))
        lrc_lines.append(f'[{from_second_str}{from_ms}]{item["content"]}')
    return '\n'.join(lrc_lines)


def get_random_available_port() -> int:
    from socket import socket

    with socket() as s:
        s.bind(('', 0))
        return s.getsockname()[1]


if __name__ == '__main__':
    print(get_random_available_port())



import json
import logging
from datetime import timedelta
from http.cookiejar import MozillaCookieJar, CookieJar
from typing import Type, Optional, Union, List
from threading import Lock

import requests.cookies
from cachetools import TTLCache
from .compat import BaseModel

from fuo_bilibili.api.audio import AudioMixin
from fuo_bilibili.api.base import BaseMixin
from fuo_bilibili.api.history import HistoryMixin
from fuo_bilibili.api.live import LiveMixin
from fuo_bilibili.api.login import LoginMixin
from fuo_bilibili.api.media import MediaMixin
from fuo_bilibili.api.playlist import PlaylistMixin
from fuo_bilibili.api.schema.requests import *  # noqa
from fuo_bilibili.api.schema.responses import *  # noqa
from fuo_bilibili.api.user import UserMixin
from fuo_bilibili.api.video import VideoMixin
from fuo_bilibili.const import PLUGIN_API_COOKIEJAR_FILE
from .wbi import encWbi

CACHE = TTLCache(50, ttl=timedelta(minutes=10).total_seconds())
logger = logging.getLogger(__name__)


class BilibiliApi(BaseMixin, VideoMixin, LoginMixin, PlaylistMixin, HistoryMixin, UserMixin, AudioMixin, LiveMixin,
                  MediaMixin):
    TIMEOUT = 10

    def __init__(self):
        self._headers = {
            'Referer': 'https://www.bilibili.com/',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2)'
                          ' AppleWebKit/537.36 (KHTML, like Gecko)'
                          ' Chrome/33.0.1750.152 Safari/537.36'
        }
        self._session = requests.Session()
        # UPDATE(2024-01-11): To make self.nav_info work, the header needs to be added
        # UPDATE(2023-xx-xx): The header cause some API failing: self.nav_info
        self._session.headers.update(self._headers)
        self._wbi: Optional[NavInfoResponse.NavInfoResponseData.Wbi] = None

        # I guess the cookie is updated by the server periodically.
        # Update the cookie file, so that it is still valid when the app is restarted.
        # HACK: since feeluown has no mechanism to update the cookie file,
        # we save cookie every N successful requests to keep it updated :)
        self.enable_auto_save_cookie = True
        self._auto_save_cookie_counter = 0
        self._auto_save_cookie_lock = Lock()

    def maybe_save_cookie(self):
        from fuo_bilibili.login import dump_user_cookies

        need_save = False
        if self.enable_auto_save_cookie:
            with self._auto_save_cookie_lock:
                self._auto_save_cookie_counter += 1
                if self._auto_save_cookie_counter >= 20:
                    need_save = True
                    self._auto_save_cookie_counter = 0
        if need_save is True:
            logger.info("auto save cookie")
            cookiedict = dict(self.get_cookiejar())
            dump_user_cookies(cookiedict)

    def get_session(self):
        return self._session

    def get_cookiejar(self):
        return self._session.cookies

    def from_cookiedict(self, cookies):
        cookiejar = requests.cookies.cookiejar_from_dict(cookies)
        self._session.cookies = cookiejar

    def set_wbi(self, wbi: NavInfoResponse.NavInfoResponseData.Wbi):
        self._wbi = wbi

    @staticmethod
    def clear_cache_by_url(urls: List[str]):
        for key in CACHE.keys():
            if len(key) < 2:
                continue
            url = key[1]
            if url in urls:
                CACHE.pop(key)

    def _get_csrf(self):
        for cookie in self.get_cookiejar():
            if cookie.name == 'bili_jct':
                return cookie.value
        raise RuntimeError('bili_jct not found')

    def get_uncached(self, url: str, param: Optional[BaseRequest],
                     clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) \
            -> Union[BaseResponse, BaseModel, None]:
        if param is None:
            r = self._session.get(url, timeout=self.TIMEOUT, **kwargs)
        else:
            def handle_base_request_param(param: BaseRequest):
                return json.loads(param.json(exclude_none=True))

            def handle_base_wbi_request_param(param: BaseWbiRequest):
                img_key = self._wbi.img_url.rsplit('/', 1)[1].split('.')[0]
                sub_key = self._wbi.sub_url.rsplit('/', 1)[1].split('.')[0]
                js = json.loads(param.json(exclude_none=True))
                return encWbi(js, img_key, sub_key)

            if isinstance(param, BaseCsrfRequest):
                param.csrf = self._get_csrf()
                js = json.loads(param.json(exclude_none=True))
            elif isinstance(param, BaseOptionalWbiRequest):
                if self._wbi is not None:
                    js = handle_base_wbi_request_param(param)
                else:
                    js = handle_base_request_param(param)
            elif isinstance(param, BaseWbiRequest):
                if self._wbi is None:
                    raise RuntimeError('wbi info is empty (not logged in)')
                js = handle_base_wbi_request_param(param)
            else:
                js = handle_base_request_param(param)
            r = self._session.get(url, timeout=self.TIMEOUT,
                                  params=js,
                                  **kwargs)
        if r.status_code != 200:
            print(r.text)
            raise RuntimeError('http not 200')
        if clazz is None:
            return None
        response_str = r.text
        logger.debug(f'api response body: {response_str}')
        res: Union[BaseResponse, BaseModel] = clazz.parse_raw(response_str)
        if isinstance(res, BaseResponse) and res.code != 0:
            raise RuntimeError(f'code not ok: {res}, url:{url}, param:{param}')
        self.maybe_save_cookie()
        return res

    def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None],
            **kwargs) \
            -> Union[BaseResponse, BaseModel, None]:
        return self.get_uncached(url, param, clazz, **kwargs)

    def get_content(self, url: str) -> str:
        return self._session.get(url).text

    def get_content_raw(self, url: str) -> bytes:
        return self._session.get(url).content

    def post(self, url: str, param: Optional[BaseRequest], clazz: Type[BaseResponse], is_json=False, **kwargs) \
            -> BaseResponse:
        if param is None:
            r = self._session.post(url, timeout=self.TIMEOUT, **kwargs)
        else:
            if isinstance(param, BaseCsrfRequest):
                param.csrf = self._get_csrf()
            request = json.loads(param.json(exclude_none=True, by_alias=True))
            if is_json:
                r = self._session.post(url, timeout=self.TIMEOUT, json=request, **kwargs)
            else:
                r = self._session.post(url, timeout=self.TIMEOUT, data=request, **kwargs)
        if r.status_code != 200:
            raise RuntimeError(f'http not 200: {r.status_code}')
        response_str = r.text
        res = clazz.parse_raw(response_str)
        if res.code != 0:
            raise RuntimeError(f'code not ok: {res.code} {res.message}')
        self.maybe_save_cookie()
        return res

    def close(self):
        try:
            self._session.close()
        except Exception as e:
            print(f'Something is wrong when destroy api connection: {str(e)}')


def main():
    from fuo_bilibili.login import load_user_cookies
    api = BilibiliApi()
    cookiedict = load_user_cookies()
    api.from_cookiedict(cookiedict)
    # info = api.history_later_videos()
    # info = api.history_add_later_videos(HistoryAddLaterVideosRequest(bvid='BV1RN4y1j7k6'))
    # info = api.video_get_info(VideoInfoRequest(bvid='BV1wnSYBQE4y'))
    # print(info)
    data = api.nav_info().data
    api.set_wbi(data.wbi_img)
    r = api.video_get_url(PlayUrlRequest(
        bvid='BV1wnSYBQE4y',
        cid=34471741344,
        fnval=VideoFnval.DASH
    ))
    print(r.data.dash.flac['audio']['baseUrl'])
    api.close()


if __name__ == '__main__':
    main()



from typing import Type, Any, Optional, Union, List

from .compat import BaseModel

from fuo_bilibili.api.schema.requests import BaseRequest, PaginatedRequest, HistoryAddLaterVideosRequest, \
    HistoryDelLaterVideosRequest
from fuo_bilibili.api.schema.responses import BaseResponse, HistoryLaterVideoResponse, HistoryVideoResponse


class HistoryMixin:
    APIX_BASE = 'https://api.bilibili.com/x'

    def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None]) -> Any:
        pass

    def post(self, url: str, param: Optional[BaseRequest], clazz: Type[BaseResponse], is_json=False, **kwargs) -> Any:
        pass

    @staticmethod
    def clear_cache_by_url(urls: List[str]):
        pass

    def history_later_videos(self) -> HistoryLaterVideoResponse:
        url = f'{self.APIX_BASE}/v2/history/toview'
        return self.get(url, None, HistoryLaterVideoResponse)

    def history_videos(self, request: PaginatedRequest) -> HistoryVideoResponse:
        url = f'{self.APIX_BASE}/v2/history'
        return self.get(url, request, HistoryVideoResponse)

    def history_add_later_videos(self, request: HistoryAddLaterVideosRequest) -> BaseResponse:
        url = f'{self.APIX_BASE}/v2/history/toview/add'
        self.clear_cache_by_url([f'{self.APIX_BASE}/v2/history/toview'])
        return self.post(url, request, BaseResponse)

    def history_del_later_videos(self, request: HistoryDelLaterVideosRequest) -> BaseResponse:
        url = f'{self.APIX_BASE}/v2/history/toview/del'
        self.clear_cache_by_url([f'{self.APIX_BASE}/v2/history/toview'])
        return self.post(url, request, BaseResponse)



from pathlib import Path

# Define data root directory
from time import time

FEELUOWN_DIRECTORY = Path.home() / '.FeelUOwn'
PLUGIN_DATA_DIRECTORY = FEELUOWN_DIRECTORY / 'fuo_bilibili'
DANMAKU_DIRECTORY = PLUGIN_DATA_DIRECTORY / 'subtitles'

# Cookiejar file
PLUGIN_API_COOKIEJAR_FILE = PLUGIN_DATA_DIRECTORY / 'bilibili_api.cookie'
PLUGIN_API_COOKIEDICT_FILE = PLUGIN_DATA_DIRECTORY / 'bilibili_api_cookie.json'

# Ensure directories
PLUGIN_DATA_DIRECTORY.mkdir(parents=True, exist_ok=True)
DANMAKU_DIRECTORY.mkdir(parents=True, exist_ok=True)

# Clear old files
for f in DANMAKU_DIRECTORY.glob('*.ass'):
    if time() - f.stat().st_mtime > 7 * 24 * 60 * 60:
        f.unlink(missing_ok=True)

for f in DANMAKU_DIRECTORY.glob('*.xml'):
    if time() - f.stat().st_mtime > 7 * 24 * 60 * 60:
        f.unlink(missing_ok=True)



from typing import Type, Any, Optional, Union

from .compat import BaseModel

from fuo_bilibili.api.schema.requests import BaseRequest, MediaGetListRequest, MediaFavlistRequest
from fuo_bilibili.api.schema.responses import BaseResponse, MediaGetListResponse, MediaFavlistResponse


class MediaMixin:
    API_PGC_BASE = 'https://api.bilibili.com/pgc'
    APIX_BASE = 'https://api.bilibili.com/x'

    def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) -> Any:
        pass

    def get_uncached(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) -> Any:
        pass

    def media_bangumi_get_list(self, request: MediaGetListRequest) -> MediaGetListResponse:
        url = f'{self.API_PGC_BASE}/view/web/season'
        return self.get(url, request, MediaGetListResponse)

    def media_bangumi_favlist(self, request: MediaFavlistRequest) -> MediaFavlistResponse:
        url = f'{self.APIX_BASE}/space/bangumi/follow/list'
        return self.get(url, request, MediaFavlistResponse)



import logging

from feeluown.gui.base_renderer import LibraryTabRendererMixin
from feeluown.gui.page_containers.table import Renderer
from feeluown.gui.widgets import TextButton
from feeluown.gui.widgets.tabbar import Tab

from fuo_bilibili import BilibiliProvider

logger = logging.getLogger(__name__)


async def render(req, **kwargs):
    app = req.ctx['app']
    app.ui.right_panel.set_body(app.ui.table_container)
    provider = app.library.get('bilibili')
    tab_id = Tab(int(req.query.get('tab_id', Tab.songs.value)))
    renderer = RankingRenderer(tab_id, provider)
    await app.ui.table_container.set_renderer(renderer)


class RankingRenderer(Renderer, LibraryTabRendererMixin):
    def __init__(self, tab_id, provider):
        self.refresh_btn = None
        self.tab_id = tab_id
        self._provider: BilibiliProvider = provider
        self._rcmd_index = 0

    async def render(self):
        self.render_tabbar()
        self.meta_widget.show()
        self.meta_widget.title = '全站排行'

        if self.tab_id == Tab.songs:
            self._refresh_hot_videos()
        if self.tab_id == Tab.playlists:
            self._refresh_weekly_playlists()

    def _refresh_hot_videos(self):
        self.show_songs(self._provider.most_popular_videos())

    def _refresh_weekly_playlists(self):
        self.show_playlists(self._provider.weekly_video_playlists())

    def show_by_tab_id(self, tab_id):
        query = {'tab_id': tab_id.value}
        self._app.browser.goto(page='/providers/bilibili/ranking', query=query)

    def render_tabbar(self):
        super().render_tabbar()
        try:
            self.tabbar.songs_btn.setText('全站热门')
            self.tabbar.albums_btn.hide()
            self.tabbar.artists_btn.hide()
            self.tabbar.videos_btn.hide()
            self.tabbar.playlists_btn.setText('每周必看')
        except Exception as e:
            logger.warning(str(e))



import time
from http.server import HTTPServer, BaseHTTPRequestHandler
from pathlib import Path
from urllib.parse import urlparse, parse_qs

from urllib3.util import parse_url
from fuo_bilibili.qt_compat import QThread


class GeetestHttpHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        url = parse_url(self.path)
        match url.path:
            case '/challenge':
                self.send_response(200, "OK")
                self.send_header("Content-Type", "text/html; charset=utf-8")
                self.end_headers()
                with (Path(__file__).parent / 'index.html').open('r') as f:
                    self.wfile.write(f.read().encode())
                    self.wfile.flush()
                return
            case '/back':
                data = parse_qs(url.query)
                validate = data['validate'][0]
                seccode = data['seccode'][0]
                challenge = data['challenge'][0]
                token = data['token'][0]
                type_ = data['type'][0]
                self.server._instance.auth_success.emit(
                    validate, seccode, challenge, token, type_)
                self.send_response(200, "OK")
                self.send_header("Content-Type", "text/plain; charset=utf-8")
                self.end_headers()
                self.wfile.write("You can safely close now!".encode())
                self.wfile.flush()
                return
        self.send_response(404, "Not Found")
        self.end_headers()


class GeetestHttpServer(HTTPServer):
    def __init__(self, address, handler, instance):
        super().__init__(address, handler)
        self._instance = instance


class GeetestAuthServer(QThread):
    def __init__(self, port: int, instance):
        super().__init__()
        self._port = port
        self._instance = instance

    def run(self) -> None:
        self._server = GeetestHttpServer(('', self._port), GeetestHttpHandler, self._instance)
        self._server.serve_forever()

    def stop(self):
        self._server.shutdown()


if __name__ == '__main__':
    thread = GeetestAuthServer(12900)
    thread.start()
    time.sleep(100000)



from typing import Type, Any, Optional, Union

from .compat import BaseModel

from fuo_bilibili.api.schema.requests import BaseRequest, PaginatedRequest, AudioFavoriteSongsRequest, \
    AudioGetUrlRequest
from fuo_bilibili.api.schema.responses import BaseResponse, AudioFavoriteListResponse, AudioFavoriteSongsResponse, \
    AudioFavoriteInfoResponse, AudioGetUrlResponse


class AudioMixin:
    API_AUDIO_BASE = 'https://www.bilibili.com/audio'

    def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) -> Any:
        pass

    def get_uncached(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) -> Any:
        pass

    def audio_favorite_list(self, request: PaginatedRequest) -> AudioFavoriteListResponse:
        url = f'{self.API_AUDIO_BASE}/music-service-c/web/collections/list'
        return self.get(url, request, AudioFavoriteListResponse)

    def audio_collected_list(self, request: PaginatedRequest) -> AudioFavoriteListResponse:
        url = f'{self.API_AUDIO_BASE}/music-service-c/web/collect/menus'
        return self.get(url, request, AudioFavoriteListResponse)

    def audio_favorite_info(self, request: AudioFavoriteSongsRequest) -> AudioFavoriteInfoResponse:
        url = f'{self.API_AUDIO_BASE}/music-service-c/web/collections/info'
        return self.get(url, request, AudioFavoriteInfoResponse)

    def audio_collected_info(self, request: AudioFavoriteSongsRequest) -> AudioFavoriteInfoResponse:
        url = f'{self.API_AUDIO_BASE}/music-service-c/web/menu/info'
        return self.get(url, request, AudioFavoriteInfoResponse)

    def audio_favorite_songs(self, request: AudioFavoriteSongsRequest) -> AudioFavoriteSongsResponse:
        url = f'{self.API_AUDIO_BASE}/music-service-c/web/song/of-coll'
        return self.get(url, request, AudioFavoriteSongsResponse)

    def audio_collected_songs(self, request: AudioFavoriteSongsRequest) -> AudioFavoriteSongsResponse:
        url = f'{self.API_AUDIO_BASE}/music-service-c/web/song/of-menu'
        return self.get(url, request, AudioFavoriteSongsResponse)

    def audio_get_url(self, request: AudioGetUrlRequest) -> AudioGetUrlResponse:
        url = f'{self.API_AUDIO_BASE}/music-service-c/web/url'
        return self.get(url, request, AudioGetUrlResponse)



from typing import Type, Any, Optional, Union

from .compat import BaseModel

from fuo_bilibili.api.schema.requests import SearchRequest, BaseRequest, PaginatedRequest, HomeRecommendVideosRequest, \
    HomeDynamicVideoRequest
from fuo_bilibili.api.schema.responses import SearchResponse, BaseResponse, NavInfoResponse, \
    HomeRecommendVideosResponse, HomeDynamicVideoResponse


class BaseMixin:
    API_BASE = 'https://api.bilibili.com/x/web-interface'
    PASSPORT_BASE = 'https://passport.bilibili.com'
    APIX_BASE = 'https://api.bilibili.com/x'

    def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None]) -> Any:
        pass

    def post(self, url: str, param: Optional[BaseRequest], clazz: Type[BaseResponse], is_json=False, **kwargs) -> Any:
        pass

    def get_uncached(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None]) -> Any:
        pass

    def search(self, request: SearchRequest) -> SearchResponse:
        url = f'{self.API_BASE}/search/type'
        return self.get(url, request, SearchResponse)

    def nav_info(self) -> NavInfoResponse:
        url = f'{self.API_BASE}/nav'
        return self.get(url, None, NavInfoResponse)

    def home_recommend_videos(self, request: HomeRecommendVideosRequest) -> HomeRecommendVideosResponse:
        url = f'{self.API_BASE}/index/top/rcmd'
        return self.get_uncached(url, request, HomeRecommendVideosResponse)

    def home_dynamic_videos(self, request: HomeDynamicVideoRequest) -> HomeDynamicVideoResponse:
        url = f'{self.APIX_BASE}/polymer/web-dynamic/v1/feed/all'
        return self.get_uncached(url, request, HomeDynamicVideoResponse)



import asyncio
import logging
import json
from pathlib import Path
from typing import Optional

from feeluown.app.gui_app import GuiApp
from feeluown.gui import ProviderUiManager
from feeluown.gui.provider_ui import AbstractProviderUi, NavBtn
from feeluown.gui.widgets.login import CookiesLoginDialog, InvalidCookies
from feeluown.library import UserModel
from feeluown.utils.aio import run_fn
from feeluown.utils.dispatch import Signal

from fuo_bilibili import __identifier__, __alias__, BilibiliProvider
from fuo_bilibili.api.schema.requests import PasswordLoginRequest, SendSmsCodeRequest, SmsCodeLoginRequest
from fuo_bilibili.api.schema.responses import RequestLoginKeyResponse
from fuo_bilibili.geetest.server import GeetestAuthServer
from fuo_bilibili.qt_compat import (
    pyqtSignal, QCloseEvent, RichText, TextBrowserInteraction,
    QDialog, QLineEdit, QVBoxLayout, QPushButton, QLabel, QFrame, QTabWidget, QMessageBox,
    QAction, QInputDialog, QWidget
)
from fuo_bilibili.util import rsa_encrypt, get_random_available_port
from fuo_bilibili.const import PLUGIN_API_COOKIEDICT_FILE
from fuo_bilibili.login import load_user_cookies, dump_user_cookies

logger = logging.getLogger(__name__)


class LoginDialog(CookiesLoginDialog):

    def __init__(self, provider: BilibiliProvider, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provider = provider

    def setup_user(self, user):
        self.provider.auth(user)

    async def user_from_cookies(self, cookies):
        if not cookies:  # is None or empty
            raise InvalidCookies('empty cookies')

        self.provider._api.from_cookiedict(cookies)
        try:
            user = await run_fn(self.provider.user_info)
        except RuntimeError as e:
            raise InvalidCookies(f'get user with cookies failed: {e}')
        return user

    def load_user_cookies(self):
        return load_user_cookies()

    def dump_user_cookies(self, _, cookies):
        return dump_user_cookies(cookies)


class BAuthDialog(QDialog):
    """
    人机验证
    """
    # signal: validate, seccode, challenge, token, type
    auth_success = pyqtSignal(str, str, str, str, str)

    def __init__(self, parent=None, provider: BilibiliProvider = None):
        super(BAuthDialog, self).__init__(parent)
        self._token = None
        self.type = None
        self._provider = provider
        self.gt_input = QLineEdit(self)
        self.challenge_input = QLineEdit(self)
        self.gt_input.setReadOnly(True)
        self.gt_input.setVisible(False)
        self.challenge_input.setReadOnly(True)
        self.challenge_input.setVisible(False)
        self.auth_link = self._create_link_label('#', '正在获取验证链接...')
        self.validate_input = QLineEdit()
        self.validate_input.setPlaceholderText('validate')
        self.seccode_input = QLineEdit()
        self.seccode_input.setPlaceholderText('seccode')
        self.hint_label = QLabel(self)
        self.hint_label.setStyleSheet('QLabel {color:orange}')
        self.authed_btn = QPushButton('完成')
        # noinspection PyUnresolvedReferences
        self.authed_btn.clicked.connect(self._check_auth_and_back)
        self._layout = QVBoxLayout(self)
        # Arrage ui layout
        self.setFixedWidth(300)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.gt_input)
        self._layout.addWidget(self.challenge_input)
        self._layout.addWidget(self.auth_link)
        # self._layout.addWidget(self.validate_input)
        # self._layout.addWidget(self.seccode_input)
        self._layout.addWidget(self.hint_label)
        # self._layout.addWidget(self.authed_btn)

    def closeEvent(self, e: QCloseEvent) -> None:
        self._server.stop()
        e.accept()

    def _check_auth_and_back(self):
        # todo: use qvalidator
        if self.validate_input.text() == '' or self.seccode_input.text() == '':
            self.hint_label.setText('请输入验证结果')
            return
        # noinspection PyUnresolvedReferences
        self.auth_success.emit(
            self.validate_input.text(), self.seccode_input.text(),
            self.challenge_input.text(), self._token, self.type)
        self.close()

    def _create_link_label(self, link: str, text: str) -> QLabel:
        auth_link = QLabel(self)
        auth_link.setText(f'<a href="{link}">{text}</a>')
        auth_link.setTextFormat(RichText)
        auth_link.setTextInteractionFlags(TextBrowserInteraction)
        auth_link.setOpenExternalLinks(True)
        return auth_link

    @staticmethod
    def exception_handler(func):
        def inner(*args, **kwargs):
            self = args[0]
            try:
                return func(*args, **kwargs)
            except Exception as e:
                QMessageBox.warning(self, '错误提示', str(e))

        return inner

    @exception_handler
    def request_challenge_params(self, type_: str):
        self.auth_link.setText(f'<a href="#">正在获取验证链接...</a>')
        self.type = type_
        self.validate_input.clear()
        self.seccode_input.clear()
        data = self._provider.request_captcha()
        geetest = data.geetest
        self.gt_input.setText(geetest.gt)
        self.challenge_input.setText(geetest.challenge)
        self._token = data.token
        port = get_random_available_port()
        self._server = GeetestAuthServer(port, self)
        self._server.start()
        self.auth_link.setText(
            f'<a href="http://127.0.0.1:{port}/challenge?gt={geetest.gt}'
            f'&challenge={geetest.challenge}&token={self._token}&type={self.type}">点击链接完成验证</a>')


class BLoginDialog(QDialog):
    login_succeed = pyqtSignal()

    def __init__(self, parent=None, provider=None, ui_manger=None):
        super(BLoginDialog, self).__init__(parent)
        self._captcha_id = None
        self._provider: BilibiliProvider = provider
        self._tab = QTabWidget(self)
        self.setFixedWidth(300)
        self.setMinimumHeight(200)
        self._tab.setFixedWidth(300)
        # 验证码登录
        self._sms_tab = QFrame(self._tab)
        self.sms_username_input = QLineEdit(self._sms_tab)
        self.sms_username_input.setPlaceholderText("手机号")
        self.sms_send_btn = QPushButton('发送验证码', self._sms_tab)
        self.sms_code_input = QLineEdit(self._sms_tab)
        self.sms_code_input.setPlaceholderText("验证码")
        self.sms_ok_btn = QPushButton('登录', self._sms_tab)
        self._sms_layout = QVBoxLayout(self._sms_tab)
        # Arrage ui layout
        self._sms_layout.setContentsMargins(0, 0, 0, 0)
        self._sms_layout.setSpacing(0)
        self._sms_layout.addWidget(self.sms_username_input)
        self._sms_layout.addWidget(self.sms_send_btn)
        self._sms_layout.addWidget(self.sms_code_input)
        self._sms_layout.addWidget(self.sms_ok_btn)
        # noinspection PyUnresolvedReferences
        self.sms_send_btn.clicked.connect(self._start_auth_code)
        # noinspection PyUnresolvedReferences
        self.sms_ok_btn.clicked.connect(self._start_sms_login)

        # 密码登录
        self._pw_tab = QFrame(self._tab)
        self.username_input = QLineEdit(self._pw_tab)
        self.username_input.setPlaceholderText("账号")
        self.pw_input = QLineEdit(self._pw_tab)
        self.pw_input.setEchoMode(QLineEdit.Password)
        self.pw_input.setPlaceholderText("密码")
        self.ok_btn = QPushButton('登录', self._pw_tab)
        self._pw_layout = QVBoxLayout(self._pw_tab)
        # Arrage ui layout
        self._pw_layout.setContentsMargins(0, 0, 0, 0)
        self._pw_layout.setSpacing(0)
        self._pw_layout.addWidget(self.username_input)
        self._pw_layout.addWidget(self.pw_input)
        self._pw_layout.addWidget(self.ok_btn)
        self._auth_dialog = BAuthDialog(self, provider)
        # noinspection PyUnresolvedReferences
        self.ok_btn.clicked.connect(self._start_auth)

        # NOTE(cosven, 2023-08-01): 这两个登录方法已经不可用了
        self._sms_tab.hide()
        self._pw_tab.hide()
        # self._tab.addTab(self._sms_tab, '验证码登录')
        # self._tab.addTab(self._pw_tab, '密码登录')

        # auth back signal
        # noinspection PyUnresolvedReferences
        self._auth_dialog.auth_success.connect(self._auth_back)

    def _finish_keyring_login(self):
        self.close()
        self.login_succeed.emit()

    def _start_auth(self):
        self._auth_dialog.setWindowTitle('密码登录验证')
        self._auth_dialog.show()
        self._auth_dialog.request_challenge_params('login')

    def _start_auth_code(self):
        self._auth_dialog.setWindowTitle('发送验证码验证')
        self._auth_dialog.show()
        self._auth_dialog.request_challenge_params('sms')

    @staticmethod
    def exception_handler(func):
        def inner(*args, **kwargs):
            self = args[0]
            try:
                return func(*args, **kwargs)
            except Exception as e:
                QMessageBox.warning(self, '错误提示', str(e))

        return inner

    @exception_handler
    def _start_sms_login(self, _):
        resp = self._provider.sms_code_login(SmsCodeLoginRequest(
            tel=self.sms_username_input.text(),
            code=self.sms_code_input.text(),
            captcha_key=self._captcha_id
        ))
        self.close()

    @exception_handler
    def _continue_password_login(self, validate: str, seccode: str, challenge: str, token: str):
        key_data = self._provider.request_key()
        encrypted_pw = self._get_encrypted_password(key_data)
        resp = self._provider.password_login(PasswordLoginRequest(
            username=self.username_input.text(),
            password=encrypted_pw,
            token=token,
            challenge=challenge,
            validate=validate,
            seccode=seccode
        ))
        if resp.data.status == 2:
            # 需要验证手机号
            QMessageBox.warning(self, '登录提示', '本次登录需要验证手机号，请使用验证码登录')
            self._tab.setCurrentIndex(0)
        else:
            self.close()

    @exception_handler
    def _continue_send_sms(self, validate: str, seccode: str, challenge: str, token: str):
        resp = self._provider.sms_send_code(SendSmsCodeRequest(
            tel=self.sms_username_input.text(),
            token=token,
            challenge=challenge,
            validate=validate,
            seccode=seccode
        ))
        self._captcha_id = resp.data.captcha_key

    def _auth_back(self, validate: str, seccode: str, challenge: str, token: str, type_: str):
        self._auth_dialog._server.stop()
        self._auth_dialog.close()
        match type_:
            case 'login':
                self._continue_password_login(validate, seccode, challenge, token)
            case 'sms':
                self._continue_send_sms(validate, seccode, challenge, token)

    def _get_encrypted_password(self, key_data: RequestLoginKeyResponse):
        salted_pw = key_data.hash + self.pw_input.text()
        return rsa_encrypt(salted_pw, key_data.key)


class BProviderUi(AbstractProviderUi):
    def __init__(self, app: 'GuiApp', provider: BilibiliProvider):
        self._app = app
        self._provider = provider
        self._login_event = Signal()
        self.login_dialog = None

    @property
    def provider(self):
        return self._provider

    @property
    def login_event(self):
        return self._login_event

    def login_or_go_home(self):
        if not self._provider.has_current_user():
            # keyring 登录
            self.login_dialog = LoginDialog(
                self._provider, 'https://bilibili.com', ['SESSDATA'])
            self.login_dialog.login_succeed.connect(self.on_login_succeed)
            self.login_dialog.autologin()
            self.login_dialog.show()
        else:
            logger.info('already logged in')
            self.login_event.emit(self, 2)

    def get_colorful_svg(self) -> str:
        return (Path(__file__).parent / 'assets' / 'icon.svg').as_posix()

    def register_pages(self, route):
        from fuo_bilibili.page_home import render as home_render
        from fuo_bilibili.page_ranking import render as ranking_render

        route('/providers/bilibili/home')(home_render)
        route('/providers/bilibili/ranking')(ranking_render)

    def list_nav_btns(self):
        return [
            NavBtn(
                icon='📺',
                text='我的首页',
                cb=lambda: self._app.browser.goto(page='/providers/bilibili/home')
            ),
            NavBtn(
                icon='🔥',
                text='全站热门',
                cb=lambda: self._app.browser.goto(page='/providers/bilibili/ranking')
            ),
        ]

    def create_playlist(self):
        name, o1 = QInputDialog.getText(self._app.ui.left_panel.playlists_header, '新建收藏夹', '收藏夹标题')
        if not o1:
            return
        desc, o2 = QInputDialog.getText(self._app.ui.left_panel.playlists_header, '新建收藏夹', '收藏夹简介')
        if not o2:
            return
        privacy, o3 = QInputDialog.getItem(self._app.ui.left_panel.playlists_header, '新建收藏夹', '歌单权限', ['公开', '私有'])
        if not o3:
            return
        try:
            self._provider.user_playlist_new(name, desc, 1 if privacy == '私有' else 0)
        except Exception as e:
            QMessageBox.warning(self._app.ui.left_panel.playlists_header, '新建收藏夹失败', str(e))

    def on_login_succeed(self):
        del self.login_dialog
        self.login_event.emit(self, 1)



from enum import Enum

from feeluown.media import Quality


class SearchType(Enum):
    """
    搜索目标类型
    """
    VIDEO = 'video'  # 视频
    BANGUMI = 'media_bangumi'  # 番剧
    MEDIA = 'media_ft'  # 影视
    LIVE = 'live'  # 直播间/主播
    LIVE_ROOM = 'live_room'  # 直播间
    LIVE_USER = 'live_user'  # 主播
    ARTICLE = 'article'  # 专栏文章
    TOPIC = 'topic'  # 话题
    BILI_USER = 'bili_user'  # 用户
    PHOTO = 'photo'  # 相册
    KETANG = 'ketang'  # 课堂


class MediaType(Enum):
    bangumi = 1
    movie = 2
    document = 3
    national_creation = 4
    series = 5
    entertainment = 7


class UserType(Enum):
    """
    用户分类
    """
    ALL = 0
    UP = 1  # UP主
    NORMAL = 2
    VERIFIED = 3  # 认证用户


class VipType(Enum):
    """
    会员类型
    """
    NO = 0
    MONTH = 1  # 月度大会员
    YEAR = 2  # 年度大会员或以上


class VideoDurationType(Enum):
    """
    视频时长筛选
    """
    ALL = 0
    LT10 = 1
    GT10LT30 = 2
    GT30LT60 = 3
    GT60 = 4


class SearchOrderType(Enum):
    """
    搜索结果排序方式
    """
    # 视频、专栏、相册
    TOTALRANK = 'totalrank'  # 综合排序
    CLICK = 'click'  # 最多点击
    PUBDATE = 'pubdate'  # 最新发布
    DM = 'dm'  # 最多弹幕
    STOW = 'stow'  # 最多收藏
    SCORES = 'scores'  # 最多评论
    # 相册
    ATTENTION = 'attention'  # 最多喜欢
    # 直播间
    ONLINE = 'online'  # 人气直播
    LIVE_TIME = 'live_time'  # 最新开播
    # 用户
    USER_DEFAULT = '0'  # 默认排序
    FANS = 'fans'  # 粉丝数
    LEVEL = 'level'  # 用户等级


class VideoQualityNum(Enum):
    """
    视频清晰度
    """
    q240 = 6
    q360 = 16
    q480 = 32
    q720 = 64
    # 以下清晰度需要登录
    q720p60 = 74
    q1080 = 80
    # 以下清晰度一般需要登录大会员账号
    q1080p = 112
    q1080p60 = 116
    q4k = 120
    qhdr = 125  # HDR 真彩色
    qdolby = 126  # 杜比视界
    q8k = 127

    def get_quality(self) -> Quality.Video:
        if self.value >= 80:
            return Quality.Video.fhd
        if self.value >= 64:
            return Quality.Video.hd
        if self.value >= 32:
            return Quality.Video.sd
        return Quality.Video.ld

    @classmethod
    def get_max_from_quality(cls, quality: Quality.Video) -> int:
        match quality:
            case Quality.Video.fhd:
                return cls.q8k.value
            case Quality.Video.hd:
                return cls.q720p60.value
            case Quality.Video.sd:
                return cls.q480.value
        return cls.q360.value


class VideoFnval(Enum):
    """
    视频获取方式
    """
    DEFAULT = 0
    FLV = 2
    MP4 = 1  # 仅240P/360P 有限速
    DASH = 16  # Dash 音视频分流
    DASH_ALT = 80


class CodecId(Enum):
    """
    CodecId
    """
    AUDIO = 0
    AVC = 7
    HEVC = 12
    AV1 = 13


class VideoCopyright(Enum):
    """
    视频类型原创或转载
    """
    ORIGIN = 1
    FORWARD = 2
    UNKNOWN = 3


class VideoState(Enum):
    PASSED = 1
    PUBLIC = 0
    PENDING_AUDIT = -1
    REJECTED = -2
    POLICE_LOCKED = -3
    LOCKED = -4
    ADMIN_LOCKED = -5
    FIXED_PENDING_AUDIT = -6
    TEMP_PAUSED_AUDIT = -7
    RESUBMIT_PENDING_AUDIT = -8
    PENDING_TRANSCODE = -9
    DELAYED_AUDIT = -10


class VideoFrom(Enum):
    VUPLOAD = 'vupload'  # 普通上传（B站）
    HUNAN = 'hunan'  # 芒果TV
    QQ = 'qq'  # 腾讯视频


class FavoriteResourceOrderType(Enum):
    MTIME = 'mtime'  # 最近收藏
    VIEW = 'view'  # 最多播放
    PUBTIME = 'pubtime'  # 最新投稿


class CommentType(Enum):
    video = 1  # 视频稿件
    topic = 2  # 话题
    activity = 3  # 活动
    audio = 14  # 音频



"""
Copied from
https://github.com/SocialSisterYi/bilibili-API-collect/blob/master/docs/misc/sign/wbi.md
"""

from functools import reduce
from hashlib import md5
import urllib.parse
import time


mixinKeyEncTab = [
    46, 47, 18, 2, 53, 8, 23, 32, 15, 50, 10, 31, 58, 3, 45, 35, 27, 43, 5, 49,
    33, 9, 42, 19, 29, 28, 14, 39, 12, 38, 41, 13, 37, 48, 7, 16, 24, 55, 40,
    61, 26, 17, 0, 1, 60, 51, 30, 4, 22, 25, 54, 21, 56, 59, 6, 63, 57, 62, 11,
    36, 20, 34, 44, 52
]

def getMixinKey(orig: str):
    '对 imgKey 和 subKey 进行字符顺序打乱编码'
    return reduce(lambda s, i: s + orig[i], mixinKeyEncTab, '')[:32]


def encWbi(params: dict, img_key: str, sub_key: str):
    '为请求参数进行 wbi 签名'
    mixin_key = getMixinKey(img_key + sub_key)
    curr_time = round(time.time())
    params['wts'] = curr_time                                   # 添加 wts 字段
    params = dict(sorted(params.items()))                       # 按照 key 重排参数
    # 过滤 value 中的 "!'()*" 字符
    params = {
        k : ''.join(filter(lambda chr: chr not in "!'()*", str(v)))
        for k, v
        in params.items()
    }
    query = urllib.parse.urlencode(params)                      # 序列化参数
    wbi_sign = md5((query + mixin_key).encode()).hexdigest()    # 计算 w_rid
    params['w_rid'] = wbi_sign
    return params
